import os.path
import platform
import shutil
import sys
import tempfile
import textwrap

import mock

from egginst.tests.common import mkdtemp
from egginst.vendor.six import StringIO
from egginst.vendor.six.moves import unittest

from enstaller.plat import custom_plat

from enstaller import __version__

from enstaller.auth import APITokenAuth, UserPasswordAuth
from enstaller.config import (prepend_url, print_config,
                              _is_using_epd_username, convert_auth_if_required,
                              _keyring_backend_name, write_default_config)
from enstaller.config import (KEYRING_SERVICE_NAME,
                              Configuration, add_url)
from enstaller.session import Session
from enstaller.errors import (EnstallerException,
                              InvalidConfiguration)
from enstaller.utils import PY_VER

from .common import (make_keyring_available_context, make_keyring_unavailable,
                     mock_print, fake_keyring_context,
                     fake_keyring, DummyAuthenticator)

FAKE_USER = "john.doe"
FAKE_PASSWORD = "fake_password"
FAKE_AUTH = UserPasswordAuth(FAKE_USER, FAKE_PASSWORD)
FAKE_CREDS = UserPasswordAuth(FAKE_USER, FAKE_PASSWORD)._encoded_auth


class TestWriteConfig(unittest.TestCase):
    def setUp(self):
        self.d = tempfile.mkdtemp()
        self.f = os.path.join(self.d, ".enstaller4rc")

    def tearDown(self):
        shutil.rmtree(self.d)

    @make_keyring_unavailable
    def test_simple(self):
        config = Configuration()
        config.update(auth=(FAKE_USER, FAKE_PASSWORD))
        config.write(self.f)

        config = Configuration.from_file(self.f)

        self.assertEqual(config.auth._encoded_auth, FAKE_CREDS)
        self.assertEqual(config.autoupdate, True)
        self.assertEqual(config.proxy, None)
        self.assertEqual(config.proxy_dict, {})
        self.assertEqual(config.use_webservice, True)
        self.assertEqual(config.webservice_entry_point,
                         "https://api.enthought.com/eggs/{0}/".format(custom_plat))
        self.assertEqual(config.api_url,
                         "https://api.enthought.com/accounts/user/info/")

    @make_keyring_unavailable
    def test_simple_with_proxy(self):
        proxystr = "http://acme.com:3128"

        config = Configuration()
        config.update(proxy=proxystr)
        config.write(self.f)

        config = Configuration.from_file(self.f)
        self.assertEqual(str(config.proxy), proxystr)
        self.assertEqual(config.proxy_dict, {"http": proxystr})

    def test_add_url(self):
        # Given
        path = os.path.join(self.d, "foo.cfg")

        config = Configuration()
        config.write(path)

        url = "http://acme.com/{PLATFORM}/"

        # When
        add_url(path, config, url)

        # Then
        with open(path, "r") as fp:
            data = fp.read()
            self.assertRegex(data, "http://acme.com")

    @make_keyring_unavailable
    def test_change_store_url(self):
        config = Configuration()
        config.update(auth=(FAKE_USER, FAKE_PASSWORD))
        config.write(self.f)

        config = Configuration.from_file(self.f)
        config.update(store_url="https://acme.com")

        self.assertEqual(config.auth._encoded_auth, FAKE_CREDS)
        self.assertEqual(config.autoupdate, True)
        self.assertEqual(config.proxy, None)
        self.assertEqual(config.use_webservice, True)
        self.assertEqual(config.webservice_entry_point,
                         "https://acme.com/eggs/{0}/".format(custom_plat))
        self.assertEqual(config.api_url,
                         "https://acme.com/accounts/user/info/")


class TestConfigKeyringConversion(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_use_epd_username(self):
        data = "EPD_auth = '{0}'".format(FAKE_CREDS)

        self.assertFalse(_is_using_epd_username(StringIO(data)))

        data = "#EPD_auth = '{0}'".format(FAKE_CREDS)
        self.assertFalse(_is_using_epd_username(StringIO(data)))

        data = textwrap.dedent("""\
            #EPD_auth = '{0}'
            EPD_username = '{1}'
        """).format(FAKE_CREDS, FAKE_USER)
        self.assertTrue(_is_using_epd_username(StringIO(data)))

    @fake_keyring
    def test_conversion(self):
        """
        Ensure we don't convert EPD_auth to using keyring.
        """
        path = os.path.join(self.prefix, "some_config_file")
        old_config = "EPD_auth = '{0}'".format(FAKE_CREDS)

        with open(path, "w") as fp:
            fp.write(old_config)

        self.assertFalse(convert_auth_if_required(path))

    @fake_keyring
    def test_username_to_auth_conversion(self):
        """
        Ensure we don't convert EPD_auth to using keyring.
        """
        # Given
        r_content = "EPD_auth = '{0}'".format(FAKE_CREDS)
        path = os.path.join(self.prefix, ".enstaller4rc")

        old_config = "EPD_username = '{0}'".format(FAKE_USER)
        with open(path, "w") as fp:
            fp.write(old_config)

        with fake_keyring_context() as mocked_keyring:
            mocked_keyring.set_password(KEYRING_SERVICE_NAME, FAKE_USER,
                                        FAKE_PASSWORD)

            # When
            converted = convert_auth_if_required(path)

            # Then
            self.assertTrue(converted)
            with open(path) as fp:
                self.assertMultiLineEqual(fp.read(), r_content)

    @fake_keyring
    def test_auth_conversion_without_password_in_keyring(self):
        # Given
        path = os.path.join(self.prefix, ".enstaller4rc")
        old_config = "EPD_username = '{0}'".format(FAKE_USER)
        with open(path, "w") as fp:
            fp.write(old_config)

        with fake_keyring_context():
            # When/Then
            with self.assertRaises(EnstallerException):
                convert_auth_if_required(path)


class TestGetAuth(unittest.TestCase):
    def setUp(self):
        self.d = tempfile.mkdtemp()
        self.f = os.path.join(self.d, ".enstaller4rc")

    def tearDown(self):
        shutil.rmtree(self.d)

    def test_with_keyring(self):
        with make_keyring_available_context() as mocked_keyring:
            config = Configuration()
            config.update(auth=(FAKE_USER, FAKE_PASSWORD))

            self.assertEqual(config.auth,
                             UserPasswordAuth(FAKE_USER, FAKE_PASSWORD))
            self.assertFalse(mocked_keyring.set_password.called)

    @make_keyring_unavailable
    def test_with_auth(self):
        config = Configuration()
        config.update(auth=(FAKE_USER, FAKE_PASSWORD))

        self.assertEqual(config.auth, FAKE_AUTH)

    @make_keyring_unavailable
    def test_without_auth_or_keyring(self):
        config = Configuration()
        self.assertIsNone(config.auth)


class TestWriteAndChangeAuth(unittest.TestCase):
    def test_simple_set_auth(self):
        # Given
        config = Configuration()

        # When/Then
        self.assertIsNone(config.auth)

        # Given
        config = Configuration()

        # When/Then
        with self.assertRaises(InvalidConfiguration):
            config.update(auth=("", ""))

        # Given
        config = Configuration()
        config.update(auth=("yoyoma", ""))

        # When/Then
        self.assertIsNotNone(config.auth)

    @make_keyring_unavailable
    def test_change_existing_config_file(self):
        r_new_password = "ouioui_dans_sa_petite_voiture"
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("EPD_auth = '{0}'".format(FAKE_CREDS))

        config = Configuration.from_file(fp.name)
        self.assertEqual(config.auth, FAKE_AUTH)

        config.update(auth=(FAKE_USER, r_new_password))
        config._change_auth(fp.name)
        new_config = Configuration.from_file(fp.name)

        self.assertEqual(new_config.auth,
                         UserPasswordAuth(FAKE_USER, r_new_password))

    @make_keyring_unavailable
    def test_change_existing_config_file_empty_username(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("EPD_auth = '{0}'".format(FAKE_CREDS))

        config = Configuration.from_file(fp.name)
        self.assertEqual(config.auth, FAKE_AUTH)

        config.reset_auth()
        config._change_auth(fp.name)

        new_config = Configuration.from_file(fp.name)
        self.assertIsNone(new_config.auth)

    def test_change_existing_config_file_without_keyring(self):
        # Given
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("EPD_auth = '{0}'".format(FAKE_CREDS))

        # When
        config = Configuration.from_file(fp.name)
        config.update(auth=("user", "dummy"))
        config._change_auth(fp.name)

        # Then
        with open(fp.name, "r") as f:
            content = f.read()
            self.assertNotRegex(content, "EPD_username")
            self.assertRegex(content, "EPD_auth")
        config = Configuration.from_file(fp.name)
        self.assertEqual(config.auth, UserPasswordAuth("user", "dummy"))

    @make_keyring_unavailable
    def test_change_empty_config_file_empty_username(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("")

        config = Configuration.from_file(fp.name)
        self.assertIsNone(config.auth)

        config.update(auth=(FAKE_USER, FAKE_PASSWORD))
        self.assertEqual(config.auth,
                         UserPasswordAuth(FAKE_USER, FAKE_PASSWORD))

    @make_keyring_unavailable
    def test_no_config_file(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("")

        config = Configuration()
        self.assertIsNone(config.auth)

        config.update(auth=(FAKE_USER, FAKE_PASSWORD))
        config.write(fp.name)

        new_config = Configuration.from_file(fp.name)
        self.assertEqual(new_config.auth, FAKE_AUTH)

    @make_keyring_unavailable
    def test_change_auth_wo_existing_auth(self):
        r_output = "EPD_auth = '{0}'\n".format(FAKE_CREDS)

        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("")

        config = Configuration.from_file(fp.name)
        config.update(auth=(FAKE_USER, FAKE_PASSWORD))
        config._change_auth(fp.name)

        with open(fp.name) as fp:
            self.assertMultiLineEqual(fp.read(), r_output)

    # FIXME: do we really want to revert the behaviour of change_auth() with
    # auth == (None, None) to do nothing ?
    @unittest.expectedFailure
    @make_keyring_unavailable
    def test_change_config_file_empty_auth(self):
        config_data = "EPD_auth = '{0}'".format(FAKE_CREDS)
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write(config_data)

        config = Configuration.from_file(fp.name)
        config.update(auth=(None, None))
        config._change_auth(fp.name)

        with open(fp.name, "r") as fp:
            self.assertEqual(fp.read(), config_data)


class TestAuthenticationConfiguration(unittest.TestCase):
    @make_keyring_unavailable
    def test_without_configuration_no_keyring(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("")

        config = Configuration.from_file(fp.name)
        self.assertIsNone(config.auth)

    def test_without_configuration_with_keyring(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            fp.write("")

        with mock.patch("enstaller.config.keyring"):
            config = Configuration.from_file(fp.name)
            self.assertIsNone(config.auth)

    @make_keyring_unavailable
    def test_with_configuration_no_keyring(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            auth_line = "EPD_auth = '{0}'".format(FAKE_CREDS)
            fp.write(auth_line)

        config = Configuration.from_file(fp.name)
        self.assertIsNotNone(config.auth)

    def test_with_configuration_with_keyring(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            auth_line = "EPD_username = '{0}'".format(FAKE_USER)
            fp.write(auth_line)

        mocked_keyring = mock.Mock(["get_password"])
        with mock.patch("enstaller.config.keyring", mocked_keyring):
            config = Configuration.from_file(fp.name)
            self.assertIsNotNone(config.auth)


class TestConfigurationParsing(unittest.TestCase):
    def test_parse_simple_unsupported_entry(self):
        # XXX: ideally, we would like something like with self.assertWarns to
        # check for the warning, but backporting the python 3.3 code to
        # unittest2 is a bit painful.
        with mock.patch("enstaller.config.warnings.warn") as m:
            Configuration.from_file(StringIO("nono = 'le petit robot'"))
            m.assert_called_with('Unsupported configuration setting nono, ignored')

    @make_keyring_unavailable
    def test_epd_auth_wo_keyring(self):
        s = StringIO("EPD_auth = '{0}'".format(FAKE_CREDS))

        config = Configuration.from_file(s)
        self.assertEqual(config.auth, FAKE_AUTH)

    @make_keyring_unavailable
    def test_epd_username_wo_keyring(self):
        """
        Ensure config auth correctly reports itself as non configured when
        using EPD_username but keyring is not available to get password.
        """
        s = StringIO("EPD_username = '{0}'".format(FAKE_USER))

        config = Configuration.from_file(s)
        self.assertIsNone(config.auth)

    def test_epd_username(self):
        with fake_keyring_context() as mocked_keyring:
            mocked_keyring.set_password(KEYRING_SERVICE_NAME, FAKE_USER, FAKE_PASSWORD)

            s = StringIO("EPD_username = '{0}'".format(FAKE_USER))
            config = Configuration.from_file(s)

            self.assertEqual(config.auth, FAKE_AUTH)
            self.assertEqual(config.auth._encoded_auth, FAKE_CREDS)

    def test_epd_auth(self):
        """
        Ensure config auth information is properly set-up when using EPD_auth
        """
        s = StringIO("EPD_auth = '{0}'".format(FAKE_CREDS))
        config = Configuration.from_file(s)

        self.assertEqual(config.auth, FAKE_AUTH)

    def test_store_kind(self):
        """
        Ensure config auth information is properly set-up when using EPD_auth
        """
        # Given
        config = Configuration()

        # When
        config.update(store_url="http://acme.com")

        # Then
        self.assertEqual(config.store_kind, "legacy")
        self.assertEqual(config.store_url, "http://acme.com")

        # Given
        config = Configuration()

        # When
        config.update(store_url="brood+http://acme.com")

        # Then
        self.assertEqual(config.store_kind, "brood")
        self.assertEqual(config.store_url, "http://acme.com")

        # Given
        s = StringIO("store_url = 'http://acme.com'")

        # When
        config = Configuration.from_file(s)

        # Then
        self.assertEqual(config.store_kind, "legacy")
        self.assertEqual(config.store_url, "http://acme.com")

        # Given
        s = StringIO("store_url = 'brood+http://acme.com'")

        # When
        config = Configuration.from_file(s)

        # Then
        self.assertEqual(config.store_kind, "brood")
        self.assertEqual(config.store_url, "http://acme.com")

    def test_api_token(self):
        # Given
        config = Configuration()

        # When
        config.update(auth=APITokenAuth("dummy token"))

        # Then
        self.assertEqual(config.auth, APITokenAuth("dummy token"))

        # Given
        data = StringIO("api_token = 'yoyo'")

        # When
        config = Configuration.from_file(data)

        # Then
        self.assertEqual(config.auth, APITokenAuth("yoyo"))

    def test_both_auth_set(self):
        # Given
        data = "EPD_auth = '{0}'\napi_token = 'token'"
        data = StringIO(data.format(FAKE_CREDS))

        msg = "Both 'EPD_auth' and 'api_token' set in configuration." \
              "\nYou should remove one of those for consistent " \
              "behaviour."

        # When
        with self.assertWarnsRegex(Warning, msg):
            Configuration.from_file(data)


class TestConfigurationPrint(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_simple_in_memory(self):
        output_template = textwrap.dedent("""\
            Python version: {pyver}
            enstaller version: {version}
            sys.prefix: {sys_prefix}
            platform: {platform}
            architecture: {arch}
            use_webservice: True
            keyring backend: {keyring_backend}
            settings:
                prefix = {{prefix}}
                repository_cache = {{repository_cache}}
                noapp = False
                proxy = None
            No valid auth information in configuration, cannot authenticate.
            You are not logged in.  To log in, type 'enpkg --userpass'.
        """).format(pyver=PY_VER, sys_prefix=os.path.normpath(sys.prefix),
                    version=__version__, platform=platform.platform(),
                    arch=platform.architecture()[0],
                    keyring_backend=_keyring_backend_name())

        config = Configuration()
        prefix = config.prefix
        repository_cache = os.path.join(prefix, "LOCAL-REPO")
        r_output = output_template.format(prefix=os.path.normpath(prefix),
                                          repository_cache=repository_cache)

        with mock_print() as m:
            print_config(config, Session(DummyAuthenticator(), self.prefix))
            self.assertMultiLineEqual(m.value, r_output)

    def test_simple(self):
        output_template = textwrap.dedent("""\
            Python version: {pyver}
            enstaller version: {version}
            sys.prefix: {sys_prefix}
            platform: {platform}
            architecture: {arch}
            use_webservice: True
            config file: {{config_file}}
            keyring backend: {keyring_backend}
            settings:
                prefix = {{prefix}}
                repository_cache = {{repository_cache}}
                noapp = False
                proxy = None
            No valid auth information in configuration, cannot authenticate.
            You are not logged in.  To log in, type 'enpkg --userpass'.
        """).format(pyver=PY_VER, sys_prefix=os.path.normpath(sys.prefix),
                    version=__version__, platform=platform.platform(),
                    arch=platform.architecture()[0],
                    keyring_backend=_keyring_backend_name())

        try:
            with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
                fp.write("")
            config = Configuration.from_file(fp.name)
        finally:
            os.unlink(fp.name)

        prefix = config.prefix
        repository_cache = os.path.join(prefix, "LOCAL-REPO")
        r_output = output_template.format(prefix=os.path.normpath(prefix),
                                          config_file=fp.name,
                                          repository_cache=repository_cache)

        with mock_print() as m:
            print_config(config, Session(DummyAuthenticator(), self.prefix))
            self.assertMultiLineEqual(m.value, r_output)

    def test_simple_no_webservice(self):
        output_template = textwrap.dedent("""\
            Python version: {pyver}
            enstaller version: {version}
            sys.prefix: {sys_prefix}
            platform: {platform}
            architecture: {arch}
            use_webservice: False
            keyring backend: {keyring_backend}
            settings:
                prefix = {{prefix}}
                repository_cache = {{repository_cache}}
                noapp = False
                proxy = None
                IndexedRepos:
                    'http://acme.com/'
            No valid auth information in configuration, cannot authenticate.
            You are not logged in.  To log in, type 'enpkg --userpass'.
        """).format(pyver=PY_VER, sys_prefix=os.path.normpath(sys.prefix),
                    version=__version__, platform=platform.platform(),
                    arch=platform.architecture()[0],
                    keyring_backend=_keyring_backend_name())

        prefix = os.path.normpath(sys.prefix)
        repository_cache = os.path.join(prefix, "LOCAL-REPO")
        r_output = output_template.format(prefix=os.path.normpath(prefix),
                                          repository_cache=repository_cache)

        config = Configuration()
        config.update(indexed_repositories=["http://acme.com"],
                      use_webservice=False)

        with mock_print() as m:
            print_config(config, Session(DummyAuthenticator(), self.prefix))
            self.assertMultiLineEqual(m.value, r_output)


class TestConfiguration(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def assertSamePath(self, left, right):
        """Very naive implementation of path comparison."""
        if sys.platform == "win32":
            self.assertEqual(left.lower(), right.lower())
        else:
            self.assertEqual(left, right)

    def test_unsupported_syntax(self):
        # Given
        data = textwrap.dedent("""\
        store_url = 'http://acme'
        store_url += '.com'
        """)
        r_message = "Could not parse configuration file (error at line 2: " \
            "expression \"store_url += '.com'\" not supported)"

        # When
        with self.assertRaises(InvalidConfiguration) as e:
            Configuration.from_file(StringIO(data))

        # Then
        self.assertMultiLineEqual(str(e.exception), r_message)

    def test_invalid_syntax(self):
        # Given
        data = "store_url = http://acme.com"
        r_message = "Could not parse configuration file (invalid python " \
                    "syntax at line 1: expression 'store_url = " \
                    "http://acme.com')"

        # When
        with self.assertRaises(InvalidConfiguration) as e:
            Configuration.from_file(StringIO(data))

        # Then
        self.assertMultiLineEqual(str(e.exception), r_message)

    def test_proxy_setup(self):
        # Given
        proxy_string = "http://acme.com"

        # When
        config = Configuration()
        config.update(proxy=proxy_string)

        # Then
        self.assertEqual(str(config.proxy), "http://acme.com:3128")

    def test_verify_ssl_setup(self):
        # When
        config = Configuration()

        # Then
        self.assertTrue(config.verify_ssl)

        # Given
        data = StringIO("verify_ssl = True")

        # When
        config = Configuration.from_file(data)

        # Then
        self.assertTrue(config.verify_ssl)

        # Given
        data = StringIO("verify_ssl = False")

        # When
        config = Configuration.from_file(data)

        # Then
        self.assertFalse(config.verify_ssl)

    def test_max_retries_setup(self):
        # When
        config = Configuration()

        # Then
        self.assertEqual(config.max_retries, 0)

        # Given
        data = StringIO("max_retries = 1")

        # When
        config = Configuration.from_file(data)

        # Then
        self.assertEqual(config.max_retries, 1)

        # Given
        data = StringIO("max_retries = 0")

        # When
        config = Configuration.from_file(data)

        # Then
        self.assertEqual(config.max_retries, 0)

        # Given
        data = StringIO("max_retries = 'a'")

        # When/Then
        with self.assertRaises(InvalidConfiguration):
            Configuration.from_file(data)

    def test_parse_simple_unsupported_entry(self):
        # XXX: ideally, we would like to check for the warning, but doing so is
        # a bit too painful as it has not been backported to unittest2
        Configuration.from_file(StringIO("nono = 'le petit robot'"))

    def test__from_legacy_locations_simple(self):
        # Given
        default_config_file = os.path.join(self.prefix, ".enstaller4rc")
        with open(default_config_file, "w") as fp:
            fp.write("store_url = \"http://acme.com\"")

        # When
        with mock.patch("enstaller.config.legacy_configuration_read_search_order",
                        return_value=[self.prefix]):
            config = Configuration._from_legacy_locations()

        # Then
        self.assertEqual(config.store_url, "http://acme.com")

    def test__from_legacy_locations_non_existing_path(self):
        # When/Then
        with mock.patch("enstaller.config.legacy_configuration_read_search_order",
                        return_value=[self.prefix]):
            with self.assertRaises(InvalidConfiguration):
                Configuration._from_legacy_locations()

    def test_reset_auth_with_keyring(self):
        with make_keyring_available_context():
            config = Configuration()
            config.update(auth=(FAKE_USER, FAKE_PASSWORD))

            config.reset_auth()

            self.assertIsNone(config.auth)

    def test_set_prefix(self):
        homedir = os.path.normpath(os.path.expanduser("~"))

        config = Configuration()
        config.update(prefix=os.path.normpath("~/.env"))

        self.assertEqual(config.prefix, os.path.join(homedir, ".env"))

    def test_set_repository_cache(self):
        homedir = os.path.normpath(os.path.expanduser("~"))

        config = Configuration()
        config.update(repository_cache="~/.env/LOCAL-REPO")
        self.assertEqual(config.repository_cache, os.path.join(homedir, ".env", "LOCAL-REPO"))

    def test_set_epd_auth(self):
        config = Configuration()
        config.set_auth_from_encoded(FAKE_CREDS)

        self.assertEqual(config.auth, FAKE_AUTH)

    def test_set_invalid_epd_auth(self):
        config = Configuration()

        with self.assertRaises(InvalidConfiguration):
            config.set_auth_from_encoded(FAKE_USER)

    def test_write_default_config_simple(self):
        # Given
        path = os.path.join(self.prefix, ".enstaller4rc")

        # When
        write_default_config(path)

        # Then
        self.assertTrue(os.path.exists(path))

    def test_write_default_config_already_exists(self):
        # Given
        path = os.path.join(self.prefix, ".enstaller4rc")
        with open(path, "w") as fp:
            fp.write("")

        # When/Then
        with self.assertRaises(EnstallerException):
            write_default_config(path)

    def test_indices_property(self):
        # Given
        r_indices = tuple([
            ("https://api.enthought.com/eggs/{0}/index.json?pypi=true".
             format(custom_plat),
             "https://api.enthought.com/eggs/{0}/index.json".format(custom_plat)),
        ])
        config = Configuration()

        # When/Then
        self.assertEqual(config.indices, r_indices)

    def test_indices_property_platform(self):
        # Given
        platform = "win-32"
        r_indices = tuple([
            ("https://api.enthought.com/eggs/{0}/index.json?pypi=true".
             format(platform),
             "https://api.enthought.com/eggs/{0}/index.json".format(platform)),
        ])

        # When
        config = Configuration()
        config._platform = platform

        # Then
        self.assertEqual(config.indices, r_indices)

        # Given
        platform = "osx-32"
        r_indices = tuple([
            ("https://api.enthought.com/eggs/{0}/index.json?pypi=true".
             format(platform),
             "https://api.enthought.com/eggs/{0}/index.json".format(platform)),
        ])

        # When
        config = Configuration()
        config._platform = "osx-32"

        # Then
        self.assertEqual(config.indices, r_indices)

    def test_indices_property_no_pypi(self):
        # Given
        r_indices = tuple([
            ("https://api.enthought.com/eggs/{0}/index.json?pypi=false".
             format(custom_plat),
             "https://api.enthought.com/eggs/{0}/index.json".format(custom_plat)),
        ])
        config = Configuration()
        config.update(use_pypi=False)

        # When/Then
        self.assertEqual(config.indices, r_indices)

    def test_indices_property_no_webservice(self):
        # Given
        r_indices = tuple((
            ("https://acme.com/{0}/index.json".format(custom_plat),
             "https://acme.com/{0}/index.json".format(custom_plat)),
        ))
        config = Configuration()
        config.update(use_webservice=False,
                      indexed_repositories=["https://acme.com/{PLATFORM}/"])

        # When/Then
        self.assertEqual(config.indices, r_indices)

    def test_from_file_complete_combination1(self):
        # Given
        if sys.platform == "win32":
            r_prefix = "C:\\tmp"
        else:
            r_prefix = "/tmp"
        r_repository_cache = r_prefix
        fp = StringIO(textwrap.dedent("""\
        EPD_auth = "{creds}"

        repository_cache = {prefix!r}
        prefix = {prefix!r}
        use_webservice = True

        store_url = "http://acme.com"
        use_pypi = False
        autoupdate = False
        noapp = False
        """.format(creds=FAKE_CREDS, prefix=r_prefix)))

        # When
        config = Configuration.from_file(fp)

        # Then
        self.assertEqual(config.auth, FAKE_AUTH)
        self.assertSamePath(config.repository_cache, r_repository_cache)
        self.assertSamePath(config.prefix, r_prefix)
        self.assertEqual(config.use_webservice, True)
        self.assertEqual(config.store_url, "http://acme.com")
        self.assertEqual(config.use_pypi, False)
        self.assertEqual(config.autoupdate, False)
        self.assertEqual(config.noapp, False)

    def test_from_file_complete_combination2(self):
        # Given
        if sys.platform == "win32":
            r_prefix = "C:\\tmp"
        else:
            r_prefix = "/tmp"
        r_repository_cache = r_prefix
        fp = StringIO(textwrap.dedent("""\
        EPD_auth = "{creds}"

        repository_cache = {prefix!r}
        prefix = {prefix!r}
        use_webservice = False

        store_url = "http://acme.com"
        use_pypi = True
        autoupdate = True
        noapp = True
        """.format(creds=FAKE_CREDS, prefix=r_prefix)))

        # When
        config = Configuration.from_file(fp)

        # Then
        self.assertEqual(config.auth, FAKE_AUTH)
        self.assertSamePath(config.repository_cache, r_repository_cache)
        self.assertSamePath(config.prefix, r_prefix)
        self.assertEqual(config.use_webservice, False)
        self.assertEqual(config.store_url, "http://acme.com")
        self.assertEqual(config.use_pypi, True)
        self.assertEqual(config.autoupdate, True)
        self.assertEqual(config.noapp, True)

    def test_setup_in_init(self):
        # Given
        store_url = "acme.com"
        max_retries = 5

        # When
        config = Configuration(store_url=store_url, max_retries=max_retries)

        # Then
        self.assertEqual(config.max_retries, max_retries)
        self.assertEqual(config.store_url, store_url)

    def test_repository_cache(self):
        # Given no value supplied, default prefix

        # When
        config = Configuration()

        # Then
        self.assertEqual(config.repository_cache,
                         os.path.normpath(os.path.join(sys.prefix,
                                                       "LOCAL-REPO")))

        # Given a prefix, but no value supplied
        prefix = self.prefix

        # When
        config = Configuration(prefix=prefix)

        # Then
        self.assertEqual(config.repository_cache, os.path.join(prefix, "LOCAL-REPO"))

        # Given a value
        repository_cache = os.path.join(self.prefix, "mycache")

        # When
        config = Configuration(repository_cache=repository_cache)

        # Then
        self.assertEqual(config.repository_cache, repository_cache)


class TestMisc(unittest.TestCase):
    def test_writable_repository_cache(self):
        config = Configuration()
        with mkdtemp() as d:
            config.update(repository_cache=d)
            self.assertEqual(config.repository_cache, d)

    def test_non_writable_repository_cache(self):
        fake_dir = "/some/dummy_dir/hopefully/doesnt/exists"

        config = Configuration()

        with mock.patch("os.makedirs", side_effect=OSError("mocked makedirs")):
            config.update(repository_cache=fake_dir)
            self.assertNotEqual(config.repository_cache, fake_dir)


class TestPrependUrl(unittest.TestCase):
    def setUp(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            pass

        self.filename = fp.name

    def tearDown(self):
        os.unlink(self.filename)

    def test_simple(self):
        r_output = textwrap.dedent("""\
            IndexedRepos = [
              'url1',
              'url2',
              'url3',
            ]
        """)

        content = textwrap.dedent("""\
            IndexedRepos = [
              'url2',
              'url3',
            ]
        """)
        with open(self.filename, "w") as fp:
            fp.write(content)

        prepend_url(self.filename, "url1")
        with open(self.filename, "r") as fp:
            self.assertMultiLineEqual(fp.read(), r_output)

    def test_invalid(self):
        content = textwrap.dedent("""\
            #IndexedRepos = [
            #  'url1',
            #]
        """)
        with open(self.filename, "w") as fp:
            fp.write(content)

        with self.assertRaises(SystemExit):
            prepend_url(self.filename, "url1")


class TestYamlConfiguration(unittest.TestCase):
    def test_default(self):
        # Given/When
        config = Configuration.from_yaml_filename(StringIO(""))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertFalse(config.indices, [])
        self.assertEqual(config.repository_cache,
                         os.path.normpath(os.path.join(sys.prefix,
                                                       "LOCAL-REPO")))
        self.assertEqual(config.store_kind, "brood")

    def test_invalid_format(self):
        # Given
        yaml_string = textwrap.dedent("""\
            repositoriess:
              - enthought/commercial
        """)

        # When/Then
        with self.assertRaises(InvalidConfiguration):
            Configuration.from_yaml_filename(StringIO(yaml_string))

    def test_load_from_path(self):
        # Given
        yaml_string = textwrap.dedent("""\
            repositories:
              - enthought/free
              - enthought/commercial
        """)
        platform = custom_plat
        r_indices = tuple((
            ('https://api.enthought.com/repo/enthought/free/{0}/index.json'.format(platform),
             'https://api.enthought.com/repo/enthought/free/{0}/index.json'.format(platform)),
            ('https://api.enthought.com/repo/enthought/commercial/{0}/index.json'.format(platform),
             'https://api.enthought.com/repo/enthought/commercial/{0}/index.json'.format(platform))
        ))

        with mkdtemp() as prefix:
            path = os.path.join(prefix, "enstaller.yaml")
            with open(path, "wt") as fp:
                fp.write(yaml_string)

            # When
            config = Configuration.from_yaml_filename(path)

            # Then
            self.assertFalse(config.use_webservice)
            self.assertEqual(config.store_url, "https://api.enthought.com")
            self.assertEqual(config.indices, r_indices)

    def test_simple(self):
        # Given
        yaml_string = textwrap.dedent("""\
            store_url: "http://acme.com"

            repositories:
              - enthought/free
              - enthought/commercial
        """)
        platform = custom_plat
        r_indices = tuple((
            ('http://acme.com/repo/enthought/free/{0}/index.json'.format(platform),
             'http://acme.com/repo/enthought/free/{0}/index.json'.format(platform)),
            ('http://acme.com/repo/enthought/commercial/{0}/index.json'.format(platform),
             'http://acme.com/repo/enthought/commercial/{0}/index.json'.format(platform))
        ))

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertEqual(config.store_url, "http://acme.com")
        self.assertEqual(config.indices, r_indices)
        self.assertEqual(config.max_retries, 0)
        self.assertTrue(config.verify_ssl)

    def test_api_token_authentication(self):
        # Given
        yaml_string = textwrap.dedent("""\
            authentication:
              kind: token
              api_token: ulysse
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertEqual(config.auth, APITokenAuth("ulysse"))

        # Given
        yaml_string = textwrap.dedent("""\
            authentication:
              api_token: ulysse
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertEqual(config.auth, APITokenAuth("ulysse"))

    def test_simple_authentication(self):
        # Given
        yaml_string = textwrap.dedent("""\
            authentication:
              kind: simple
              username: nono@acme.com
              password: ulysse
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertEqual(config.auth,
                         UserPasswordAuth("nono@acme.com", "ulysse"))

    def test_basic_authentication(self):
        # Given
        yaml_string = textwrap.dedent("""\
            authentication:
              kind: basic
              auth: {0}
        """.format(UserPasswordAuth("nono@acme.com", "ulysse")._encoded_auth))

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertEqual(config.auth,
                         UserPasswordAuth("nono@acme.com", "ulysse"))

    def test_files_cache(self):
        # Given
        yaml_string = textwrap.dedent("""\
            files_cache: "/foo/bar"
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertEqual(config.repository_cache, "/foo/bar")

        # Given
        yaml_string = textwrap.dedent("""\
            files_cache: "~/foo/bar/{PLATFORM}"
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.use_webservice)
        self.assertEqual(config.repository_cache,
                         os.path.expanduser("~/foo/bar/{0}".format(custom_plat)))

    def test_max_retries(self):
        # Given
        yaml_string = textwrap.dedent("""\
            max_retries: 1
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertEqual(config.max_retries, 1)

    def test_verify_ssl(self):
        # Given
        yaml_string = textwrap.dedent("""\
            verify_ssl: True
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertTrue(config.verify_ssl)

        # Given
        yaml_string = textwrap.dedent("""\
            verify_ssl: False
        """)

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertFalse(config.verify_ssl)

    def test_repositories(self):
        # Given
        yaml_string = textwrap.dedent("""\
            store_url: "http://www.acme.com"

            repositories:
              - "enthought/free"
              - "file:///foo"
        """)
        r_repositories = (
            "http://www.acme.com/repo/enthought/free/{0}/".format(custom_plat),
            "file:///foo/".format(custom_plat),
        )

        # When
        config = Configuration.from_yaml_filename(StringIO(yaml_string))

        # Then
        self.assertEqual(config.indexed_repositories, r_repositories)



from __future__ import absolute_import

import re
import shutil
import sys
import tempfile
import textwrap
import zlib

import mock

from egginst._compat import assertCountEqual
from egginst.main import EggInst
from egginst.tests.common import DUMMY_EGG, mkdtemp
from egginst.vendor.six.moves import unittest

from enstaller.config import Configuration
from enstaller.enpkg import Enpkg
from enstaller.repository import Repository
from enstaller.session import Session
from enstaller.tests.common import (DummyAuthenticator, FakeOptions,
                                    create_prefix_with_eggs,
                                    dummy_installed_package_factory,
                                    dummy_repository_package_factory,
                                    mocked_session_factory,
                                    mock_index, mock_print, mock_raw_input,
                                    PY_VER)
from enstaller.vendor import requests, responses
from enstaller.versions.enpkg import EnpkgVersion

from ..utils import (disp_store_info, exit_if_root_on_non_owned, install_req,
                     install_time_string, name_egg, print_installed,
                     repository_factory, updates_check)
from ..utils import _fetch_json_with_progress, _print_warning


if sys.version_info < (2, 7):
    # FIXME: this looks quite fishy. On 2.6, with unittest2, the assertRaises
    # context manager does not contain the actual exception object ?
    def exception_code(ctx):
        return ctx.exception
else:
    def exception_code(ctx):
        return ctx.exception.code


FAKE_AUTH = ("nono", "le gros robot")
V = EnpkgVersion.from_string


class TestExitIfRootOnNonOwned(unittest.TestCase):
    @mock.patch("enstaller.cli.utils.sys.platform", "win32")
    def test_windows(self):
        exit_if_root_on_non_owned()

    @unittest.skipIf(sys.platform == "win32", "no getuid on windows")
    @mock.patch("enstaller.cli.utils.sys.platform", "linux")
    @mock.patch("os.getuid", lambda: 0)
    def test_sudo(self):
        with mock.patch("enstaller.cli.utils.is_running_on_non_owned_python",
                        return_value=False):
            exit_if_root_on_non_owned()

        with mock.patch("enstaller.cli.utils.is_running_on_non_owned_python",
                        return_value=True):
            with mock_raw_input(""):
                with self.assertRaises(SystemExit):
                    exit_if_root_on_non_owned()

            with mock_raw_input("yes"):
                exit_if_root_on_non_owned()


class TestMisc(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test__print_warning(self):
        # Given
        message = "fail to crash"
        r_output = "Warning: fail to crash\n\n"

        # When
        with mock_print() as m:
            _print_warning(message)

        # Then
        self.assertMultiLineEqual(m.value, r_output)

    def test_disp_store_info(self):
        store_location = "https://api.enthought.com/eggs/osx-64/"
        self.assertEqual(disp_store_info(store_location), "api osx-64")

        store_location = "https://api.enthought.com/eggs/win-32/"
        self.assertEqual(disp_store_info(store_location), "api win-32")

        self.assertEqual(disp_store_info(""), "-")

    def test_name_egg(self):
        name = "foo-1.0.0-1.egg"
        self.assertEqual(name_egg(name), "foo")

        name = "fu_bar-1.0.0-1.egg"
        self.assertEqual(name_egg(name), "fu_bar")

        with self.assertRaises(AssertionError):
            name = "some/dir/fu_bar-1.0.0-1.egg"
            name_egg(name)

    @responses.activate
    def test_repository_factory(self):
        self.maxDiff = None

        # Given
        store_url = "https://acme.com"
        config = Configuration(store_url=store_url, use_webservice=False)
        config.set_repositories_from_names(["enthought/foo"])

        responses.add(responses.GET, config.indices[0][0],
                      status=404)

        session = mocked_session_factory(self.tempdir)
        r_warning = textwrap.dedent("""\
            Warning: Could not fetch the following indices:

                     - 'neko'

                     Those repositories do not exist (or you do not have the rights to
                     access them). You should edit your configuration to remove those
                     repositories.

        """)

        # When
        with mock_print() as m:
            with mock.patch("enstaller.cli.utils._display_store_name",
                            return_value="neko"):
                repository = repository_factory(session, config.indices)

        # Then
        self.assertEqual(len(list(repository.iter_packages())), 0)
        self.assertMultiLineEqual(m.value, r_warning)

        # When/Then
        with self.assertRaises(requests.exceptions.HTTPError):
            repository_factory(session, config.indices, raise_on_error=True)


class TestInfoStrings(unittest.TestCase):
    def test_print_install_time(self):
        with mkdtemp():
            installed_entries = [dummy_installed_package_factory("dummy",
                                                                 "1.0.1", 1)]
            installed_repository = Repository()
            for package in installed_entries:
                installed_repository.add_package(package)

            self.assertRegexpMatches(install_time_string(installed_repository,
                                                         "dummy"),
                                     "dummy-1.0.1-1.egg was installed on:")

            self.assertEqual(install_time_string(installed_repository,
                                                 "ddummy"),
                             "")

    def test_print_installed(self):
        with mkdtemp() as d:
            r_out = textwrap.dedent("""\
                Name                 Version              Store
                ============================================================
                dummy                1.0.1-1              {0}
                """.format(disp_store_info(d)))
            ec = EggInst(DUMMY_EGG, d)
            ec.install()

            repository = Repository._from_prefixes([d])
            with mock_print() as m:
                print_installed(repository)
            self.assertMultiLineEqual(m.value, r_out)

            r_out = textwrap.dedent("""\
                Name                 Version              Store
                ============================================================
                """)

            repository = Repository._from_prefixes([d])
            with mock_print() as m:
                print_installed(repository, pat=re.compile("no_dummy"))
            self.assertEqual(m.value, r_out)


class TestUpdatesCheck(unittest.TestCase):
    def _create_repositories(self, entries, installed_entries):
        repository = Repository()
        for entry in entries:
            repository.add_package(entry)

        installed_repository = Repository()
        for entry in installed_entries:
            installed_repository.add_package(entry)

        return repository, installed_repository

    def test_update_check_new_available(self):
        # Given
        remote_entries = [
            dummy_repository_package_factory("dummy", "1.2.0", 1),
            dummy_repository_package_factory("dummy", "0.9.8", 1)
        ]
        installed_entries = [
            dummy_installed_package_factory("dummy", "1.0.1", 1)
        ]

        remote_repository, installed_repository = \
            self._create_repositories(remote_entries, installed_entries)

        # When
        updates, EPD_update = updates_check(remote_repository,
                                            installed_repository)

        # Then
        self.assertEqual(EPD_update, [])
        self.assertEqual(len(updates), 1)
        update0 = updates[0]
        assertCountEqual(self, update0.keys(), ["current", "update"])
        self.assertEqual(update0["current"].version, V("1.0.1-1"))
        self.assertEqual(update0["update"].version, V("1.2.0-1"))

    def test_update_check_no_new_available(self):
        # Given
        remote_entries = [
            dummy_repository_package_factory("dummy", "1.0.0", 1),
            dummy_repository_package_factory("dummy", "0.9.8", 1)
        ]
        installed_entries = [
            dummy_installed_package_factory("dummy", "1.0.1", 1)
        ]

        remote_repository, installed_repository = \
            self._create_repositories(remote_entries, installed_entries)

        # When
        updates, EPD_update = updates_check(remote_repository,
                                            installed_repository)

        # Then
        self.assertEqual(EPD_update, [])
        self.assertEqual(updates, [])

    def test_update_check_no_available(self):
        # Given
        installed_entries = [
            dummy_installed_package_factory("dummy", "1.0.1", 1)
        ]

        remote_repository, installed_repository = \
            self._create_repositories([], installed_entries)

        # When
        updates, EPD_update = updates_check(remote_repository,
                                            installed_repository)

        # Then
        self.assertEqual(EPD_update, [])
        self.assertEqual(updates, [])

    def test_update_check_epd(self):
        # Given
        remote_entries = [dummy_repository_package_factory("EPD", "7.3", 1)]
        installed_entries = [dummy_installed_package_factory("EPD", "7.2", 1)]

        remote_repository, installed_repository = \
            self._create_repositories(remote_entries, installed_entries)

        # When
        updates, EPD_update = updates_check(remote_repository,
                                            installed_repository)

        # Then
        self.assertEqual(updates, [])
        self.assertEqual(len(EPD_update), 1)

        epd_update0 = EPD_update[0]
        assertCountEqual(self, epd_update0.keys(), ["current", "update"])
        self.assertEqual(epd_update0["current"].version, V("7.2-1"))
        self.assertEqual(epd_update0["update"].version, V("7.3-1"))


class TestInstallReq(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_install_not_available(self):
        # Given
        config = Configuration()
        config.update(auth=FAKE_AUTH)

        nose = dummy_repository_package_factory("nose", "1.3.0", 1)
        nose.available = False
        repository = Repository()
        repository.add_package(nose)

        enpkg = Enpkg(repository,
                      mocked_session_factory(config.repository_cache),
                      [self.prefix])
        enpkg.execute = mock.Mock()

        # When/Then
        with mock.patch("enstaller.config.subscription_message") as \
                subscription_message:
            with self.assertRaises(SystemExit) as e:
                install_req(enpkg, config, "nose", FakeOptions())
            subscription_message.assert_called()
            self.assertEqual(e.exception.code, 1)

    def test_simple_install(self):
        remote_entries = [
            dummy_repository_package_factory("nose", "1.3.0", 1)
        ]

        with mock.patch("enstaller.main.Enpkg.execute") as m:
            enpkg = create_prefix_with_eggs(Configuration(), self.prefix, [],
                                            remote_entries)
            install_req(enpkg, Configuration(), "nose", FakeOptions())
            m.assert_called_with([('fetch', 'nose-1.3.0-1.egg'),
                                  ('install', 'nose-1.3.0-1.egg')])

    def test_simple_non_existing_requirement(self):
        config = Configuration()
        r_error_string = "No egg found for requirement 'nono_le_petit_robot'.\n"
        non_existing_requirement = "nono_le_petit_robot"

        with mock.patch("enstaller.main.Enpkg.execute") as mocked_execute:
            enpkg = create_prefix_with_eggs(config, self.prefix, [])
            with mock_print() as mocked_print:
                with self.assertRaises(SystemExit) as e:
                    install_req(enpkg, config, non_existing_requirement,
                                FakeOptions())
                self.assertEqual(exception_code(e), 1)
                self.assertEqual(mocked_print.value, r_error_string)
            mocked_execute.assert_not_called()

    def test_simple_no_install_needed(self):
        installed_entries = [
            dummy_installed_package_factory("nose", "1.3.0", 1)
        ]
        remote_entries = [
            dummy_repository_package_factory("nose", "1.3.0", 1)
        ]
        config = Configuration()

        with mock.patch("enstaller.main.Enpkg.execute") as m:
            enpkg = create_prefix_with_eggs(config, self.prefix,
                                            installed_entries, remote_entries)
            install_req(enpkg, config, "nose", FakeOptions())
            m.assert_called_with([])

    def test_recursive_install_unavailable_dependency(self):
        config = Configuration()
        session = Session(DummyAuthenticator(), self.prefix)

        auth = ("nono", "le gros robot")
        session.authenticate(auth)
        config.update(auth=auth)

        r_output = textwrap.dedent("""
        Cannot install 'scipy', as this package (or some of its requirements) are not
        available at your subscription level 'Canopy / EPD Free' (You are logged in as
        'nono').
        """)

        self.maxDiff = None
        numpy = dummy_repository_package_factory("numpy", "1.7.1", 1)
        numpy.available = False
        scipy = dummy_repository_package_factory("scipy", "0.12.0", 1)
        scipy.packages = ["numpy 1.7.1"]

        remote_entries = [numpy, scipy]

        with mock.patch("enstaller.main.Enpkg.execute"):
            enpkg = create_prefix_with_eggs(config, self.prefix, [], remote_entries)
            with mock_print() as m:
                with self.assertRaises(SystemExit):
                    install_req(enpkg, config, "scipy", FakeOptions())
                self.assertMultiLineEqual(m.value, r_output)

    @mock_index({
        "rednose-0.2.3-1.egg": {
            "available": True,
            "build": 1,
            "md5": "41640f27172d248ccf6dcbfafe53ba4d",
            "mtime": 1300825734.0,
            "name": "rednose",
            "packages": [],
            "product": "pypi",
            "python": PY_VER,
            "size": 9227,
            "type": "egg",
            "version": "0.2.3"
        }
    })
    def test_install_pypi_requirement(self):
        self.maxDiff = None

        # Given
        r_message = textwrap.dedent("""\
        The following packages/requirements are coming from the PyPi repo:

        rednose

        The PyPi repository which contains >10,000 untested ("as is")
        packages. Some packages are licensed under GPL or other licenses
        which are prohibited for some users. Dependencies may not be
        provided. If you need an updated version or if the installation
        fails due to unmet dependencies, the Knowledge Base article
        Installing external packages into Canopy Python
        (https://support.enthought.com/entries/23389761) may help you with
        installing it.

        Are you sure that you wish to proceed?  (y/[n])
        """)

        config = Configuration()
        session = Session.from_configuration(config)
        session.authenticate(("nono", "le petit robot"))
        repository = repository_factory(session, config.indices)

        enpkg = Enpkg(repository, session, [self.prefix])
        enpkg.execute = mock.Mock()

        # When
        with mock_print() as mocked_print:
            with mock_raw_input("yes"):
                install_req(enpkg, config, "rednose", FakeOptions())

        # Then
        self.assertMultiLineEqual(mocked_print.value, r_message)

    @mock_index({
        "rednose-0.2.3-1.egg": {
            "available": True,
            "build": 1,
            "md5": "41640f27172d248ccf6dcbfafe53ba4d",
            "mtime": 1300825734.0,
            "name": "rednose",
            "packages": ["python_termstyle"],
            "product": "pypi",
            "python": PY_VER,
            "size": 9227,
            "type": "egg",
            "version": "0.2.3"
        }
    })
    def test_install_broken_pypi_requirement(self):
        self.maxDiff = None

        # Given
        r_message = textwrap.dedent("""
Broken pypi package 'rednose-0.2.3-1.egg': missing dependency 'python_termstyle'

Pypi packages are not officially supported. If this package is important to
you, please contact Enthought support to request its inclusion in our
officially supported repository.

In the mean time, you may want to try installing 'rednose-0.2.3-1.egg' from sources
with pip as follows:

    $ enpkg pip
    $ pip install <requested_package>

""")

        config = Configuration()
        session = Session.from_configuration(config)
        session.authenticate(("nono", "le petit robot"))
        repository = repository_factory(session, config.indices)

        enpkg = Enpkg(repository, session, [self.prefix])
        enpkg.execute = mock.Mock()

        # When
        with self.assertRaises(SystemExit):
            with mock_print() as mocked_print:
                with mock_raw_input("yes"):
                    install_req(enpkg, config, "rednose", FakeOptions())

        # Then
        self.assertMultiLineEqual(mocked_print.value, r_message)

    def _assert_ask_for_pypi(self, mocked_print):
        line = "The following packages/requirements are coming from the " \
               "PyPi repo:"
        self.assertTrue(mocked_print.value.startswith(line))

    def _assert_dont_ask_for_pypi(self, mocked_print):
        line = "The following packages/requirements are coming from the " \
               "PyPi repo:"
        self.assertFalse(mocked_print.value.startswith(line))

    @mock_index({
        "swig-2.0.2-1.egg": {
            "available": True,
            "build": 1,
            "md5": "41640f27172d248ccf6dcbfafe53ba4d",
            "mtime": 1300825734.0,
            "name": "swig",
            "packages": [],
            "product": "pypi",
            "python": PY_VER,
            "size": 9227,
            "type": "egg",
            "version": "2.0.2"
        },
        "swig-2.0.12-1.egg": {
            "available": True,
            "build": 1,
            "md5": "41640f27172d248ccf6dcbfafe53ba4d",
            "mtime": 1300825734.0,
            "name": "swig",
            "packages": [],
            "product": "free",
            "python": PY_VER,
            "size": 9227,
            "type": "egg",
            "version": "2.0.12"
        }
    })
    def test_install_non_pypi_after_pypi(self):
        # Given
        config = Configuration()
        config.update(auth=("nono", "le petit robot"))

        session = Session.from_configuration(config)
        session.authenticate(config.auth)

        repository = repository_factory(session, config.indices)

        enpkg = Enpkg(repository, session, [self.prefix])
        enpkg.execute = mock.Mock()

        # When
        with mock_print() as mocked_print:
            with mock_raw_input("yes"):
                install_req(enpkg, config, "swig", FakeOptions())

        # Then
        self._assert_dont_ask_for_pypi(mocked_print)

    @mock_index({
        "swig-2.0.2-1.egg": {
            "available": True,
            "build": 1,
            "md5": "41640f27172d248ccf6dcbfafe53ba4d",
            "mtime": 1300825734.0,
            "name": "swig",
            "packages": [],
            "product": "pypi",
            "python": PY_VER,
            "size": 9227,
            "type": "egg",
            "version": "2.0.2"
        },
        "swig-2.0.12-1.egg": {
            "available": True,
            "build": 1,
            "md5": "41640f27172d248ccf6dcbfafe53ba4d",
            "mtime": 1300825734.0,
            "name": "swig",
            "packages": [],
            "product": "free",
            "python": PY_VER,
            "size": 9227,
            "type": "egg",
            "version": "2.0.12"
        }
    })
    def test_install_pypi_before_non_pypi(self):
        # Given
        config = Configuration()
        config.update(auth=("nono", "le petit robot"))

        session = Session.from_configuration(config)
        session.authenticate(config.auth)

        repository = repository_factory(session, config.indices)

        enpkg = Enpkg(repository, session, [self.prefix])
        enpkg.execute = mock.Mock()

        # When
        with mock_print() as mocked_print:
            with mock_raw_input("yes"):
                install_req(enpkg, config, "swig 2.0.2", FakeOptions())

        # Then
        self._assert_ask_for_pypi(mocked_print)

        # When
        with mock_print() as mocked_print:
            with mock_raw_input("yes"):
                install_req(enpkg, config, "swig 2.0.2-1", FakeOptions())

        # Then
        self._assert_ask_for_pypi(mocked_print)


class TestFetchJsonWithProgress(unittest.TestCase):
    def _gzip_compress(self, data):
        compressor = zlib.compressobj(6, zlib.DEFLATED, 16 + zlib.MAX_WBITS)
        body = compressor.compress(data)
        body += compressor.flush()
        return body

    @responses.activate
    def test_simple(self):

        # Given
        def callback(request):
            self.assertTrue("gzip" in
                            request.headers.get("Accept-Encoding", ""))

            headers = {"Content-Encoding": "gzip"}
            body = self._gzip_compress(b"{}")
            return (200, headers, body)

        responses.add_callback(responses.GET, "https://acme.com/index.json", callback)

        config = Configuration()

        # When
        session = Session.from_configuration(config)
        resp = session.fetch("https://acme.com/index.json")
        data = _fetch_json_with_progress(resp, "acme.com", quiet=False)

        # Then
        self.assertEqual(data, {})

    @responses.activate
    def test_handle_stripped_header(self):

        # Given
        def callback(request):
            self.assertTrue("gzip" in
                            request.headers.get("Accept-Encoding", ""))

            headers = {"Content-Encoding": ""}
            body = self._gzip_compress(b"{}")
            return (200, headers, body)

        responses.add_callback(responses.GET, "https://acme.com/index.json", callback)

        config = Configuration()

        # When
        session = Session.from_configuration(config)
        resp = session.fetch("https://acme.com/index.json")
        data = _fetch_json_with_progress(resp, "acme.com", quiet=False)

        # Then
        self.assertEqual(data, {})

    @responses.activate
    def test_handle_stripped_header_incomplete_data(self):

        # Given
        def callback(request):
            self.assertTrue("gzip" in
                            request.headers.get("Accept-Encoding", ""))

            headers = {"Content-Encoding": ""}
            incomplete_body = self._gzip_compress(b"{}")[:-1]
            return (200, headers, incomplete_body)

        responses.add_callback(responses.GET, "https://acme.com/index.json", callback)

        config = Configuration()

        # When/Then
        session = Session.from_configuration(config)
        resp = session.fetch("https://acme.com/index.json")
        with self.assertRaises(requests.exceptions.ContentDecodingError):
            _fetch_json_with_progress(resp, "acme.com", quiet=False)



import hashlib
import os.path
import shutil
import tempfile

from egginst.vendor.six.moves import unittest

from enstaller.errors import InvalidChecksum
from enstaller.fetch_utils import (MD5File, checked_content)
from enstaller.utils import compute_md5


class TestMD5File(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _write_content(self, filename, data):
        with open(filename, "wb") as fp:
            fp.write(data)

    def test_simple(self):
        # Given
        source = os.path.join(self.tempdir, "source.data")
        self._write_content(source, b"data")

        # When
        target = os.path.join(self.tempdir, "target.data")
        with open(target, "wb") as _fp:
            fp = MD5File(_fp)
            fp.write(b"data")

        # Then
        self.assertEqual(fp.checksum, compute_md5(target))
        self.assertEqual(compute_md5(target), compute_md5(source))


class TestCheckedContent(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _write_content(self, filename, data):
        with open(filename, "wb") as fp:
            fp.write(data)

    def test_simple(self):
        # Given
        data = b"data"
        checksum = hashlib.md5(data).hexdigest()
        path = os.path.join(self.tempdir, "foo.data")

        # When/Then
        with checked_content(path, checksum) as fp:
            fp.write(data)

    def test_invalid_checksum(self):
        # Given
        data = b"data"
        checksum = hashlib.md5(data).hexdigest()
        path = os.path.join(self.tempdir, "foo.data")

        # When/Then
        with self.assertRaises(InvalidChecksum):
            with checked_content(path, checksum) as fp:
                fp.write(b"")

    def test_abort(self):
        # Given
        data = b"data"
        checksum = hashlib.md5(data).hexdigest()
        path = os.path.join(self.tempdir, "foo.data")

        # When/Then
        with checked_content(path, checksum) as fp:
            fp.abort()



import functools

from enstaller.vendor.requests.adapters import HTTPAdapter

from .controller import CacheController
from .cache import DictCache
from .filewrapper import CallbackFileWrapper


class CacheControlAdapter(HTTPAdapter):
    invalidating_methods = set(['PUT', 'DELETE'])

    def __init__(self, cache=None,
                 cache_etags=True,
                 controller_class=None,
                 serializer=None,
                 heuristic=None,
                 *args, **kw):
        super(CacheControlAdapter, self).__init__(*args, **kw)
        self.cache = cache or DictCache()
        self.heuristic = heuristic

        controller_factory = controller_class or CacheController
        self.controller = controller_factory(
            self.cache,
            cache_etags=cache_etags,
            serializer=serializer,
        )

    def send(self, request, **kw):
        """
        Send a request. Use the request information to see if it
        exists in the cache and cache the response if we need to and can.
        """
        if request.method == 'GET':
            cached_response = self.controller.cached_request(request)
            if cached_response:
                return self.build_response(request, cached_response, from_cache=True)

            # check for etags and add headers if appropriate
            request.headers.update(self.controller.conditional_headers(request))

        resp = super(CacheControlAdapter, self).send(request, **kw)

        return resp

    def build_response(self, request, response, from_cache=False):
        """
        Build a response by making a request or using the cache.

        This will end up calling send and returning a potentially
        cached response
        """
        if not from_cache and request.method == 'GET':

            # apply any expiration heuristics
            if response.status == 304:
                # We must have sent an ETag request. This could mean
                # that we've been expired already or that we simply
                # have an etag. In either case, we want to try and
                # update the cache if that is the case.
                cached_response = self.controller.update_cached_response(
                    request, response
                )

                if cached_response is not response:
                    from_cache = True

                # We are done with the server response, read a
                # possible response body (compliant servers will
                # not return one, but we cannot be 100% sure) and
                # release the connection back to the pool.
                response.read(decode_content=False)
                response.release_conn()

                response = cached_response
            else:
                # Check for any heuristics that might update headers
                # before trying to cache.
                if self.heuristic:
                    response = self.heuristic.apply(response)

                # Wrap the response file with a wrapper that will cache the
                #   response when the stream has been consumed.
                response._fp = CallbackFileWrapper(
                    response._fp,
                    functools.partial(
                        self.controller.cache_response,
                        request,
                        response,
                    )
                )

        resp = super(CacheControlAdapter, self).build_response(
            request, response
        )

        # See if we should invalidate the cache.
        if request.method in self.invalidating_methods and resp.ok:
            cache_url = self.controller.cache_url(request.url)
            self.cache.delete(cache_url)

        # Give the request a from_cache attr to let people use it
        resp.from_cache = from_cache

        return resp

    def close(self):
        self.cache.close()
        super(CacheControlAdapter, self).close()



import logging
import os.path

import egginst._compat

from egginst.main import EggInst, get_installed, name_version_fn  # noqa

logging.getLogger("egginst").addHandler(egginst._compat.NullHandler())

POST_INSTALL_LIB = os.path.join(os.path.dirname(__file__), "_post_install_lib.py")



import os
import os.path
import subprocess
import textwrap
import zipfile

from setuptools import setup

from distutils.util import convert_path

from setuptools.command.bdist_egg import bdist_egg as old_bdist_egg

from egginst.main import BOOTSTRAP_ARCNAME


MAJOR = 4
MINOR = 8
MICRO = 7

IS_RELEASED = True

VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

BOOTSTRAP_SCRIPT = os.path.join(os.path.dirname(__file__), "scripts", "bootstrap.py")


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        git_revision = out.strip().decode('ascii')
    except OSError:
        git_revision = "Unknown"

    try:
        out = _minimal_ext_cmd(['git', 'rev-list', '--count', 'HEAD'])
        git_count = out.strip().decode('ascii')
    except OSError:
        git_count = "0"

    return git_revision, git_count


def write_version_py(filename='enstaller/_version.py'):
    template = """\
# THIS FILE IS GENERATED FROM ENSTALLER SETUP.PY
version = '{version}'
full_version = '{full_version}'
git_revision = '{git_revision}'
is_released = {is_released}

if not is_released:
    version = full_version
"""
    # This is only used for upgrade test. Kids, never use this by yourself !
    force_mode = os.environ.get("__TEST_FORCE_ENSTALLER_VERSION", None)
    if force_mode is not None:
        version = full_version = force_mode
        is_released = True
    else:
        version = full_version = VERSION
        is_released = IS_RELEASED

    git_rev = "Unknown"
    git_count = "0"

    if os.path.exists('.git'):
        git_rev, git_count = git_version()
    elif os.path.exists('enstaller/_version.py'):
        # must be a source distribution, use existing version file
        try:
            from enstaller._version import git_revision as git_rev
        except ImportError:
            raise ImportError("Unable to import git_revision. Try removing "
                              "enstaller/_version.py and the build directory "
                              "before building.")

    if not is_released:
        full_version += '.dev' + git_count

    with open(filename, "wt") as fp:
        fp.write(template.format(version=version,
                                 full_version=full_version,
                                 git_revision=git_rev,
                                 is_released=is_released))


class bdist_enegg(old_bdist_egg):
    def finalize_options(self):
        old_bdist_egg.finalize_options(self)

        basename = "enstaller-{0}-1.egg".format(__version__)
        self.egg_output = os.path.join(self.dist_dir, basename)

    def _write_bootstrap_code(self, bootstrap_code):
        zp = zipfile.ZipFile(self.egg_output, "a",
                             compression=zipfile.ZIP_DEFLATED)
        try:
            zp.writestr(BOOTSTRAP_ARCNAME, bootstrap_code)
        finally:
            zp.close()

    def _write_spec_depend(self, spec_depend):
        zp = zipfile.ZipFile(self.egg_output, "a",
                             compression=zipfile.ZIP_DEFLATED)
        try:
            zp.writestr("EGG-INFO/spec/depend", spec_depend)
        finally:
            zp.close()

    def run(self):
        old_bdist_egg.run(self)

        spec_depend = textwrap.dedent("""\
            metadata_version = '1.1'
            name = 'enstaller'
            version = '{0}'
            build = 1

            arch = None
            platform = None
            osdist = None
            python = None
            packages = []
        """.format(__version__))
        self._write_spec_depend(spec_depend)

        with open(BOOTSTRAP_SCRIPT, "rt") as fp:
            self._write_bootstrap_code(fp.read())


write_version_py()

kwds = {}  # Additional keyword arguments for setup

d = {}
exec(compile(open(convert_path('enstaller/__init__.py')).read(),
             convert_path('enstaller/__init__.py'),
             'exec'),
     d)
kwds['version'] = __version__ = d['__version__']

f = open('README.rst')
kwds['long_description'] = f.read()
f.close()

include_testing = True

packages = [
    'egginst',
    'egginst.console',
    'egginst.vendor',
    'enstaller',
    'enstaller.auth',
    'enstaller.cli',
    'enstaller.indexed_repo',
    'enstaller.solver',
    'enstaller.tools',
    'enstaller.vendor',
    'enstaller.vendor.cachecontrol',
    'enstaller.vendor.cachecontrol.caches',
    'enstaller.vendor.futures',
    'enstaller.vendor.jsonschema',
    'enstaller.vendor.keyring',
    'enstaller.vendor.keyring.backends',
    'enstaller.vendor.keyring.util',
    'enstaller.vendor.requests',
    'enstaller.vendor.requests.packages',
    'enstaller.vendor.requests.packages.chardet',
    'enstaller.vendor.requests.packages.urllib3',
    'enstaller.vendor.requests.packages.urllib3.contrib',
    'enstaller.vendor.requests.packages.urllib3.packages',
    'enstaller.vendor.requests.packages.urllib3.packages.ssl_match_hostname',
    'enstaller.vendor.requests.packages.urllib3.util',
    'enstaller.vendor.sqlite_cache',
    'enstaller.vendor.win32ctypes',
    'enstaller.vendor.yaml',
    'enstaller.vendor.yaml_py3',
    'enstaller.versions',
]

package_data = {"enstaller.vendor.requests": ["cacert.pem"],
                "enstaller.vendor.jsonschema": ["schemas/draft3.json", "schemas/draft4.json"]}

if include_testing:
    packages += [
        'egginst.tests',
        'enstaller.auth.tests',
        'enstaller.cli.tests',
        'enstaller.indexed_repo.tests',
        'enstaller.solver.tests',
        'enstaller.tests',
        'enstaller.tools.tests',
        'enstaller.versions.tests',
    ]
    macho_binaries = """dummy_with_target_dat-1.0.0-1.egg  foo_amd64
    foo_legacy_placehold.dylib  foo_rpath.dylib  foo.so  foo_x86
    libfoo.dylib""".split()

    package_data["egginst.tests"] = ["data/*egg", "data/zip_with_softlink.zip"]
    package_data["egginst.tests"] += [os.path.join("data", "macho", p)
                                      for p in macho_binaries]

    package_data["enstaller.indexed_repo.tests"] = [
        "*.txt",
        "epd/*.txt", "gpl/*.txt",
        "open/*.txt",
        "runner/*.txt",
    ]

setup(
    name="enstaller",
    author="Enthought, Inc.",
    author_email="info@enthought.com",
    url="https://github.com/enthought/enstaller",
    license="BSD",
    description="Install and managing tool for egg-based packages",
    packages=packages,
    package_data=package_data,
    entry_points={
        "console_scripts": [
            "enpkg = enstaller.main:main_noexc",
            "egginst = egginst.main:main",
            "enpkg-repair = egginst.repair_broken_egg_info:main",
            "update-patches = enstaller.patch:main",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Topic :: System :: Software Distribution",
        "Topic :: System :: Systems Administration",
    ],
    test_suite="nose.collector",
    cmdclass={"bdist_enegg": bdist_enegg},
    **kwds
)



from __future__ import absolute_import

import logging

import egginst._compat


logging.getLogger("enstaller").addHandler(egginst._compat.NullHandler())

try:
    from enstaller._version import (full_version as __version__,
                                    git_revision as __git_revision__,
                                    is_released as __is_released__,
                                    )
except ImportError as e:  # pragma: no cover
    __version__ = __git_revision__ = "no-built"
    __is_released__ = False

from enstaller.config import Configuration
from enstaller.repository import Repository, RepositoryPackageMetadata
from enstaller.session import Session

__all__ = ["Configuration", "Repository", "RepositoryPackageMetadata",
           "Session"]



import logging

from os.path import isfile, join

from egginst.utils import compute_md5, makedirs

from enstaller.fetch_utils import checked_content
from enstaller.repository import egg_name_to_name_version


logger = logging.getLogger(__name__)


class _CancelableResponse(object):
    def __init__(self, path, package_metadata, fetcher, force):
        self._path = path
        self._package_metadata = package_metadata

        self._canceled = False

        self._fetcher = fetcher
        self._force = force

    def cancel(self):
        self._canceled = True

    def __iter__(self):
        return self.iter_content()

    def iter_content(self):
        if not self.needs_to_download:
            return

        with checked_content(self._path, self._package_metadata.md5) as target:
            url = self._package_metadata.source_url
            response = self._fetcher.fetch(url)

            for chunk in response.iter_content(1024):
                if self._canceled:
                    response.close()
                    target.abort()
                    return

                target.write(chunk)
                yield chunk

    @property
    def needs_to_download(self):
        needs_to_download = True

        if isfile(self._path):
            if self._force:
                if compute_md5(self._path) == self._package_metadata.md5:
                    logger.info("Not refetching, %r MD5 match", self._path)
                    needs_to_download = False
            else:
                logger.info("Not forcing refetch, %r exists", self._path)
                needs_to_download = False

        return needs_to_download


class _DownloadManager(object):
    def __init__(self, url_fetcher, repository, auth=None):
        self._repository = repository
        self._fetcher = url_fetcher
        self.cache_directory = url_fetcher.cache_directory

        makedirs(self.cache_directory)

    def _path(self, fn):
        return join(self.cache_directory, fn)

    def iter_fetch(self, egg, force=False):
        """ Fetch the given egg using streaming.

        Parameters
        ----------
        egg : str
            An egg filename (e.g. 'numpy-1.8.0-1.egg')
        force : bool
            If force is True, will download even if the file is already in the
            download cache.

        Example
        -------
        Simple usage::

            downloader = DownloadManager(...)
            response = downloader.iter_fetch(egg)
            for chunk in response:
                pass

        Note
        ----
        Iterating over the response already writes the file at the usual
        location. This is mostly useful when you want to control cancellation
        and/or follow download progress.
        """
        name, version = egg_name_to_name_version(egg)
        package_metadata = self._repository.find_package(name, version)

        path = self._path(package_metadata.key)

        return _CancelableResponse(path, package_metadata, self._fetcher,
                                   force)

    def fetch(self, egg, force=False):
        """ Fetch the given egg.

        Parameters
        ----------
        egg : str
            An egg filename (e.g. 'numpy-1.8.0-1.egg')
        force : bool
            If force is True, will download even if the file is already in the
            download cache.
        """
        context = self.iter_fetch(egg, force)
        for _ in context:
            pass



from enstaller.egg_meta import split_eggname
from enstaller.errors import EnpkgError
from enstaller.repository import egg_name_to_name_version

from .resolve import Resolve


class Solver(object):
    def __init__(self, remote_repository, top_installed_repository,
                 mode='recur', force=False, forceall=False):
        self._remote_repository = remote_repository
        self._top_installed_repository = top_installed_repository

        self.mode = mode
        self.force = force
        self.forceall = forceall

    def resolve(self, request):
        operations = []

        for job in request.jobs:
            if job.kind == "install":
                operations.extend(self._install(job.requirement))
            elif job.kind == "remove":
                operations.extend(("remove", p) for p in
                                  self._remove(job.requirement))
            else:
                raise ValueError("Unsupported job kind: {0}".format(job.kind))

        return operations

    def _install(self, requirement):
        eggs = Resolve(self._remote_repository).install_sequence(requirement,
                                                                 self.mode)
        return self._install_actions(eggs, self.mode, self.force,
                                     self.forceall)

    def _remove(self, requirement):
        if requirement.version and requirement.build:
            full_version = "{0}-{1}".format(requirement.version,
                                            requirement.build)
        else:
            full_version = None
        packages = self._top_installed_repository.find_packages(
            requirement.name, full_version)
        if len(packages) == 0:
            raise EnpkgError("package %s not installed" % (requirement, ))
        return [packages[0].key]

    def _install_actions(self, eggs, mode, force, forceall):
        if not self.forceall:
            # remove already installed eggs from egg list
            if self.force:
                eggs = self._filter_installed_eggs(eggs[:-1]) + [eggs[-1]]
            else:
                eggs = self._filter_installed_eggs(eggs)

        # remove packages with the same name (from first egg collection
        # only, in reverse install order)
        res = []
        for egg in reversed(eggs):
            name = split_eggname(egg)[0].lower()
            installed_packages = self._top_installed_repository.find_packages(name)
            assert len(installed_packages) < 2
            if len(installed_packages) == 1:
                installed_package = installed_packages[0]
                res.append(('remove', installed_package.key))
        for egg in eggs:
            res.append(('install', egg))
        return res

    def _filter_installed_eggs(self, eggs):
        """ Filter out already installed eggs from the given egg list.

        Parameters
        ----------
        eggs: seq
            List of egg filenames
        """
        filtered_eggs = []
        for egg in eggs:
            name, _ = egg_name_to_name_version(egg)
            for installed in self._top_installed_repository.find_packages(name):
                if installed.key == egg:
                    break
            else:
                filtered_eggs.append(egg)
        return filtered_eggs



class _Job(object):
    def __init__(self, requirement, job_type):
        self.requirement = requirement

        if job_type not in ["install", "remove", "update", "upgrade"]:
            raise ValueError("invalid job type {0}".format(job_type))
        self.kind = job_type

    def __eq__(self, other):
        if other is None:
            return False
        else:
            return self.kind == other.kind \
                and self.requirement == other.requirement


class Request(object):
    def __init__(self):
        self.jobs = []

    def _add_job(self, requirement, job_type):
        self.jobs.append(_Job(requirement, job_type))

    def install(self, requirement):
        self._add_job(requirement, "install")

    def remove(self, requirement):
        self._add_job(requirement, "remove")



from __future__ import absolute_import, print_function

import ast
import contextlib
import errno
import hashlib
import logging
import sys
import os
import shutil
import stat
import tempfile
import zipfile

from os.path import basename, isdir, isfile, islink, join

from egginst.errors import InvalidFormat
from egginst.vendor.six import PY2, string_types

from ._zipfile import ZIP_SOFTLINK_ATTRIBUTE_MAGIC

on_win = bool(sys.platform == 'win32')

if on_win:
    bin_dir_name = 'Scripts'
    rel_site_packages = r'Lib\site-packages'
else:
    bin_dir_name = 'bin'
    rel_site_packages = 'lib/python%i.%i/site-packages' % sys.version_info[:2]

logger = logging.getLogger(__name__)


def rm_empty_dir(path):
    """
    Remove the directory `path` if it is a directory and empty.
    If the directory does not exist or is not empty, do nothing.
    """
    try:
        os.rmdir(path)
    except OSError:  # directory might not exist or not be empty
        pass


def rm_rf(path):
    if not on_win and islink(path):
        # Note that we have to check if the destination is a link because
        # exists('/path/to/dead-link') will return False, although
        # islink('/path/to/dead-link') is True.
        logger.info("Removing: %r (link)", path)
        os.unlink(path)

    elif isfile(path):
        logger.info("Removing: %r (file)", path)
        if on_win:
            try:
                os.unlink(path)
            except (WindowsError, IOError):
                os.rename(path, join(tempfile.mkdtemp(), basename(path)))
        else:
            os.unlink(path)

    elif isdir(path):
        logger.info("Removing: %r (directory)", path)
        if on_win:
            try:
                shutil.rmtree(path)
            except (WindowsError, IOError):
                os.rename(path, join(tempfile.mkdtemp(), basename(path)))
        else:
            shutil.rmtree(path)


def get_executable(prefix):
    if on_win:
        paths = [prefix, join(prefix, bin_dir_name)]
        for path in paths:
            executable = join(path, 'python.exe')
            if isfile(executable):
                return executable
    else:
        path = join(prefix, bin_dir_name, 'python')
        if isfile(path):
            from subprocess import Popen, PIPE
            cmd = [path, '-c', 'import sys;print(sys.executable)']
            p = Popen(cmd, stdout=PIPE)
            executable = p.communicate()[0].strip()
            if not PY2:
                return executable.decode()
            else:
                return executable
    return sys.executable


def human_bytes(n):
    """
    Return the number of bytes n in more human readable form.
    """
    if n < 1024:
        return '%i B' % n
    k = (n - 1) / 1024 + 1
    if k < 1024:
        return '%i KB' % k
    return '%.2f MB' % (float(n) / (2 ** 20))


def makedirs(path):
    """Recursive directory creation function that does not fail if the
    directory already exists."""
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def ensure_dir(path):
    """
    Create the parent directory of the give path, recursively is necessary.
    """
    makedirs(os.path.dirname(path))


def is_zipinfo_dir(zip_info):
    """Returns True if the given zip_info refers to a directory."""
    return stat.S_ISDIR(zip_info.external_attr >> 16)


def zip_write_symlink(fp, link_name, source):
    """Add to the zipfile the given link_name as a softlink to source

    Parameters
    ----------
    fp: file object
        ZipFile instance
    link_name: str
        Path of the symlink
    source: str
        Path the symlink points to (the output of os.readlink)
    """
    zip_info = zipfile.ZipInfo(link_name)
    zip_info.create_system = 3
    zip_info.external_attr = ZIP_SOFTLINK_ATTRIBUTE_MAGIC
    fp.writestr(zip_info, source)


def zip_has_arcname(zp, arcname):
    """
    Returns True if the given zipfile instance contains the given archive

    Parameters
    ----------
    zp: ZipFile
        The zip archive to consider
    arcname: str
        The archive to look for
    """
    try:
        zp.getinfo(arcname)
        return True
    except KeyError:
        return False


class _AssignmentParser(ast.NodeVisitor):
    def __init__(self):
        self._data = {}

    def parse(self, s):
        self._data.clear()

        root = ast.parse(s)
        self.visit(root)
        return self._data

    def generic_visit(self, node):
        if type(node) != ast.Module:
            raise InvalidFormat("Unexpected expression @ line {0}".
                                format(node.lineno), node.lineno)
        super(_AssignmentParser, self).generic_visit(node)

    def visit_Assign(self, node):
        try:
            value = ast.literal_eval(node.value)
        except ValueError:
            msg = "Invalid configuration syntax at line {0}".format(node.lineno)
            raise InvalidFormat(msg, node.lineno)
        else:
            for target in node.targets:
                self._data[target.id] = value


def parse_assignments(file_or_filename):
    """
    Parse files which contain only python assignements, and returns the
    corresponding dictionary name: value

    Parameters
    ----------
    file_or_filename: str, file object
        If a string, interpreted as a filename. File object otherwise.
    """
    if isinstance(file_or_filename, string_types):
        with open(file_or_filename) as fp:
            return _AssignmentParser().parse(fp.read())
    else:
        return _AssignmentParser().parse(file_or_filename.read())


def compute_md5(path, block_size=256 * 1024):
    """Compute the md5 checksum of the given path.

    Avoids reading the whole file in RAM, and computes the md5 in chunks.

    Parameters
    ----------
    path: str or file object
        If a string, assumed to be the path to the file to be checksumed. If a
        file object, checksum will start at the current file position.
    block_size: int
        Block size to use when reading data.
    """
    m = hashlib.md5()

    def _compute_checksum(fp):
        while True:
            data = fp.read(block_size)
            m.update(data)
            if len(data) < block_size:
                break
        return m.hexdigest()

    if isinstance(path, string_types):
        with open(path, "rb") as fp:
            return _compute_checksum(fp)
    else:
        return _compute_checksum(path)


def rename(source, target):
    if sys.platform == "win32":
        try:
            os.rename(source, target)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            else:
                os.unlink(target)
                os.rename(source, target)
    else:
        return os.rename(source, target)


@contextlib.contextmanager
def atomic_file(filename, mode='w+b'):
    """
    Context manager that allows to write data safely.

    It is safe in the sense that any error happening while writing will not
    leave a stalled file, and the target file is only accessible in the
    filesystem when no error occured.

    Parameters
    ----------
    filename: str
        Target path to write to
    mode: str
        open mode.

    Returns
    -------
    fp: file object-like
        An object with the write method available.

    Example
    -------

    The typical usage is::

        with atomic_file("foo.txt", "w") as fp:
            fp.write("some string")
            # The file 'foo.txt' does NOT exist at this point
            raise ValueError("some error is happening")
        # If no error occured, the file would exist at this point

    Note
    ----
    The atomicity is only guaranteed on posix platforms, assuming the
    filesystem supports atomic rename. If filename already exists before
    entering the context manager, the file is guaranteed to be unchanged if any
    error occured within the context manager.

    A future version may use win32 API to support atomicity on this platform as
    well.
    """

    class _FileWrapper(object):
        def __init__(self, fp):
            self._fp = fp
            # Make the name private to prevent people from re-opening the file
            # (not supported on e.g. windows, and will interfere with the
            # atomic_rename feature anyway).
            self._name = fp.name

            self._is_aborted = False

        def abort(self):
            self._is_aborted = True

        def write(self, data):
            self._fp.write(data)

    def _cleanup(temp_fp):
        if temp_fp is not None:
            try:
                os.unlink(temp_fp._name)
            except OSError as e:
                if not e.errno != errno.EEXIST:
                    raise

    temp_prefix = os.path.basename(filename)
    temp_dir = os.path.dirname(filename)

    temp_fp = None
    try:
        with tempfile.NamedTemporaryFile(
                prefix=temp_prefix, suffix='.tmp', dir=temp_dir,
                delete=False, mode=mode) as _temp_fp:
            temp_fp = _FileWrapper(_temp_fp)
            yield temp_fp
    except:
        _cleanup(temp_fp)
        raise
    else:
        if temp_fp._is_aborted:
            _cleanup(temp_fp)
        else:
            rename(temp_fp._name, filename)

if sys.platform == "win32":
    from egginst._win32_compat import samefile
else:
    samefile = os.path.samefile



import os.path
import shutil
import tempfile

import mock

from egginst.vendor.six.moves import unittest
from egginst.tests.common import NOSE_1_2_1, NOSE_1_3_0

from ..unpack import main


class TestMain(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    @mock.patch("enstaller.tools.unpack.extract_egg")
    def test_simple(self, extract_egg):
        # Given
        argv = [self.prefix, NOSE_1_3_0]

        # When
        main(argv)

        # Then
        extract_egg.assert_called_with(NOSE_1_3_0, self.prefix)

    @mock.patch("enstaller.tools.unpack.extract_egg")
    def test_multiple_eggs(self, extract_egg):
        # Given
        argv = [self.prefix, NOSE_1_2_1, NOSE_1_3_0]
        r_calls = [
            mock.call(NOSE_1_2_1, self.prefix),
            mock.call(NOSE_1_3_0, self.prefix),
        ]

        # When
        main(argv)

        # Then
        extract_egg.assert_has_calls(r_calls)

    def test_real(self):
        # Given
        argv = [self.prefix, NOSE_1_2_1]
        r_man = os.path.join(self.prefix, "share", "man", "man1",
                             "nosetests.1")

        # When
        main(argv)

        # Then
        self.assertTrue(os.path.exists(r_man))



import mock

import egginst.progress
from egginst.progress import console_progress_manager_factory
from egginst.console.simple import AFTER_BAR, BEFORE_BAR
from egginst.vendor.six.moves import unittest


DEFAULT_TERMINAL_SIZE = 80


class TestProgressBar(unittest.TestCase):
    @mock.patch("egginst.progress.get_terminal_size",
                return_value=(DEFAULT_TERMINAL_SIZE, 25))
    @mock.patch("egginst.console.simple.get_terminal_size",
                return_value=(DEFAULT_TERMINAL_SIZE, 25))
    def test_dont_overflow(self, *a):
        # Ensure we don't overflow the terminal size

        # Given
        message = "fetching"
        filename = "MKL-10.3.1-1.egg"
        size = 1024 ** 2 * 70

        # When
        with mock.patch("sys.stdout") as mocked_stdout:
            progress = console_progress_manager_factory(message,
                                                        filename, size,
                                                        show_speed=True)

            with progress:
                for _ in range(int(size / 1024)):
                    progress.update(1024)

        # Then
        for i, (args, kw) in enumerate(mocked_stdout.write.call_args_list):
            line = args[0].rstrip(AFTER_BAR).lstrip(BEFORE_BAR)
            self.assertTrue(len(line) < DEFAULT_TERMINAL_SIZE)

    def test_get_terminal_size_with_env(self):
        terminal_size = DEFAULT_TERMINAL_SIZE, 25
        with mock.patch.dict('os.environ',
                             COLUMNS=str(terminal_size[0]),
                             LINES=str(terminal_size[1])):
            self.assertEqual(egginst.progress.get_terminal_size(),
                             terminal_size)



import os.path
import shutil
import tempfile

import mock

from egginst.progress import console_progress_manager_factory, ProgressBar
from egginst.tests.common import mkdtemp, DUMMY_EGG, _EGGINST_COMMON_DATA
from egginst.utils import compute_md5, makedirs
from egginst.vendor.six.moves import unittest

from enstaller.config import Configuration
from enstaller.enpkg import Enpkg, FetchAction, InstallAction, RemoveAction
from enstaller.errors import EnpkgError, InvalidChecksum
from enstaller.fetch import _DownloadManager
from enstaller.repository import (egg_name_to_name_version,
                                  PackageMetadata, Repository,
                                  RepositoryPackageMetadata)
from enstaller.session import Session
from enstaller.utils import PY_VER, path_to_uri
from enstaller.vendor import responses

from .common import (dummy_repository_package_factory,
                     mocked_session_factory,
                     mock_history_get_state_context, repository_factory,
                     unconnected_enpkg_factory, DummyAuthenticator)


class TestEnpkgActions(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_empty_actions(self):
        """
        Ensuring enpkg.execute([]) does not crash
        """
        # Given
        enpkg = unconnected_enpkg_factory()

        # When/Then
        enpkg.execute([])

    def test_remove(self):
        with mkdtemp() as d:
            makedirs(d)

            enpkg = unconnected_enpkg_factory([d])

            with mock.patch("enstaller.enpkg.RemoveAction.execute") as mocked_remove:
                actions = [("remove", DUMMY_EGG)]
                enpkg.execute(actions)
                self.assertTrue(mocked_remove.called)

    def _prepared_mocked_egginst(self, nfiles):
        installed_size = 1024 ** 2 * 2

        remover = mock.Mock()
        remover.installed_size = installed_size

        files = ["foo{0}".format(i) for i in range(nfiles)]
        remover.files = files
        remover.remove_iterator.return_value = iter(files)

        return remover

    @mock.patch("enstaller.enpkg.EggInst")
    def test_progress(self, EggInst):
        # Ensure the accumulated arg of updates is exactly the number of files
        # to be removed

        # Given
        nfiles = 245
        remover = self._prepared_mocked_egginst(nfiles)

        EggInst.return_value._egginst_remover = remover
        top_repository = mock.Mock()
        remote_repository = mock.Mock()

        def factory(*a, **kw):
            return console_progress_manager_factory("fetching", *a, **kw)

        action = RemoveAction("yoyo-1.1-1.egg", self.prefix, top_repository,
                              remote_repository, factory)

        # When
        with mock.patch("egginst.progress.ProgressBar",
                        auto_spec=ProgressBar) as MockedProgressBar:
                for step in action.iter_execute():
                    action.progress_update(step)

        # Then
        accumulated = sum(args[0] for args, kw in
                          MockedProgressBar.return_value.update.call_args_list)
        self.assertEqual(accumulated, nfiles)


class TestEnpkgExecute(unittest.TestCase):
    def setUp(self):
        self.prefixes = [tempfile.mkdtemp()]

    def tearDown(self):
        for prefix in self.prefixes:
            shutil.rmtree(prefix)

    def test_simple_fetch(self):
        egg = "yoyo.egg"
        fetch_opcode = 0

        with mock.patch("enstaller.enpkg.FetchAction") as mocked_fetch:
            enpkg = unconnected_enpkg_factory(prefixes=self.prefixes)
            enpkg.ec = mock.MagicMock()
            enpkg.execute([("fetch_{0}".format(fetch_opcode), egg)])

            self.assertTrue(mocked_fetch.called)
            mocked_fetch.assert_called()

    def test_simple_install(self):
        config = Configuration()
        entries = [
            dummy_repository_package_factory("dummy", "1.0.1", 1)
        ]

        repository = repository_factory(entries)

        with mock.patch("enstaller.enpkg.FetchAction.execute") as mocked_fetch:
            with mock.patch("enstaller.enpkg.InstallAction.execute") as mocked_install:
                enpkg = Enpkg(repository,
                              mocked_session_factory(config.repository_cache),
                              prefixes=self.prefixes)
                actions = [("install", "dummy-1.0.1-1.egg")]
                enpkg.execute(actions)

                mocked_fetch.assert_called()
                mocked_install.assert_called_with()


class TestEnpkgRevert(unittest.TestCase):
    def setUp(self):
        self.prefixes = [tempfile.mkdtemp()]

    def tearDown(self):
        for prefix in self.prefixes:
            shutil.rmtree(prefix)

    def test_empty_history(self):
        enpkg = unconnected_enpkg_factory(self.prefixes)
        enpkg.revert_actions(0)

        with self.assertRaises(EnpkgError):
            enpkg.revert_actions(1)

    def test_invalid_argument(self):
        repository = Repository()

        enpkg = Enpkg(repository,
                      mocked_session_factory(Configuration().repository_cache),
                      prefixes=self.prefixes)
        with self.assertRaises(EnpkgError):
            enpkg.revert_actions([])

    def test_revert_missing_unavailable_egg(self):
        egg = "non_existing_dummy_egg-1.0.0-1.egg"
        enpkg = Enpkg(Repository(),
                      mocked_session_factory(Configuration().repository_cache),
                      prefixes=self.prefixes)
        with self.assertRaises(EnpkgError):
            enpkg.revert_actions(set([egg]))

    @responses.activate
    def test_simple_scenario(self):
        egg = DUMMY_EGG
        r_actions = {1: [], 0: [("remove", os.path.basename(egg))]}
        config = Configuration()

        repository = Repository()
        package = RepositoryPackageMetadata.from_egg(egg)
        package.python = PY_VER
        repository.add_package(package)

        with open(egg, "rb") as fp:
            responses.add(responses.GET, package.source_url,
                          body=fp.read(), status=200,
                          content_type='application/json')

        session = Session(DummyAuthenticator(), config.repository_cache)

        enpkg = Enpkg(repository, session, prefixes=self.prefixes)
        actions = [("fetch", os.path.basename(egg)),
                   ("install", os.path.basename(egg))]
        enpkg.execute(actions)

        name, version = egg_name_to_name_version(egg)
        enpkg._installed_repository.find_package(name, version)

        for state in [0, 1]:
            actions = enpkg.revert_actions(state)
            self.assertEqual(actions, r_actions[state])

    def test_enstaller_not_removed(self):
        enstaller_egg = set(["enstaller-4.6.2-1.egg"])
        installed_eggs = set(["dummy-1.0.0-1.egg", "another_dummy-1.0.0-1.egg"])

        with mock_history_get_state_context(installed_eggs | enstaller_egg):
            enpkg = unconnected_enpkg_factory()
            ret = enpkg.revert_actions(installed_eggs)

            self.assertEqual(ret, [])

    def test_same_state(self):
        installed_eggs = ["dummy-1.0.0-1.egg", "another_dummy-1.0.0-1.egg"]

        with mock_history_get_state_context(installed_eggs):
            enpkg = unconnected_enpkg_factory()
            ret = enpkg.revert_actions(set(installed_eggs))

            self.assertEqual(ret, [])

    def test_superset(self):
        """
        Ensure installed eggs not in current state are removed.
        """
        installed_eggs = ["dummy-1.0.0-1.egg", "another_dummy-1.0.0-1.egg"]

        with mock_history_get_state_context(installed_eggs):
            enpkg = unconnected_enpkg_factory()
            ret = enpkg.revert_actions(set(installed_eggs[:1]))

            self.assertEqual(ret, [("remove", "another_dummy-1.0.0-1.egg")])

    def test_subset(self):
        r_actions = [("fetch_0", "another_dummy-1.0.0-1.egg"),
                     ("install", "another_dummy-1.0.0-1.egg")]

        installed_eggs = ["dummy-1.0.0-1.egg"]
        revert_eggs = ["dummy-1.0.0-1.egg", "another_dummy-1.0.0-1.egg"]

        with mock_history_get_state_context(installed_eggs):
            enpkg = unconnected_enpkg_factory()
            with mock.patch.object(enpkg, "_remote_repository"):
                ret = enpkg.revert_actions(set(revert_eggs))
                self.assertEqual(ret, r_actions)


class TestFetchAction(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _downloader_factory(self, paths):
        repository = Repository()
        for path in paths:
            package = RepositoryPackageMetadata.from_egg(path)
            repository.add_package(package)

        return (_DownloadManager(Session(DummyAuthenticator(), self.tempdir), repository),
                repository)

    def _add_response_for_path(self, path):
        with open(path, "rb") as fp:
            responses.add(responses.GET, path_to_uri(path),
                          body=fp.read(), status=200,
                          content_type='application/octet-stream')

    @responses.activate
    def test_simple(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)
        downloader, repository = self._downloader_factory([path])
        self._add_response_for_path(path)

        # When
        action = FetchAction(path, downloader, repository)
        action.execute()

        # Then
        target = os.path.join(downloader.cache_directory, filename)
        self.assertTrue(os.path.exists(target))
        self.assertEqual(compute_md5(target), compute_md5(path))
        self.assertFalse(action.is_canceled)

    @responses.activate
    def test_iteration(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)
        downloader, repository = self._downloader_factory([path])
        self._add_response_for_path(path)

        # When
        action = FetchAction(path, downloader, repository)
        for step in action:
            pass

        # Then
        target = os.path.join(downloader.cache_directory, filename)
        self.assertTrue(os.path.exists(target))
        self.assertEqual(compute_md5(target), compute_md5(path))
        self.assertFalse(action.is_canceled)

    @responses.activate
    def test_iteration_cancel(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)
        downloader, repository = self._downloader_factory([path])
        self._add_response_for_path(path)

        # When
        action = FetchAction(path, downloader, repository)
        for step in action:
            action.cancel()

        # Then
        target = os.path.join(downloader.cache_directory, filename)
        self.assertFalse(os.path.exists(target))
        self.assertTrue(action.is_canceled)

    @responses.activate
    def test_progress_manager_iter(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)
        downloader, repository = self._downloader_factory([path])
        self._add_response_for_path(path)

        # When/Then
        class MyDummyProgressBar(object):
            def update(self, n):
                self.fail("progress bar called")

            def __enter__(self):
                return self

            def __exit__(self, *a, **kw):
                pass

        progress = MyDummyProgressBar()
        action = FetchAction(path, downloader, repository,
                             progress_bar_factory=lambda *a, **kw: progress)
        for step in action:
            pass

    @responses.activate
    def test_progress_manager(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)
        downloader, repository = self._downloader_factory([path])
        self._add_response_for_path(path)

        class MyDummyProgressBar(object):
            def __init__(self):
                self.called = False

            def update(self, n):
                self.called = True

            def __enter__(self):
                return self

            def __exit__(self, *a, **kw):
                pass

        progress = MyDummyProgressBar()
        action = FetchAction(path, downloader, repository,
                             progress_bar_factory=lambda *a, **kw: progress)
        action.execute()

        # Then
        self.assertTrue(progress.called)

    def _add_failing_checksum_response(self, url, payload, n_invalid=2):
        counter = [0]

        def request_callback(request):
            counter[0] += 1

            headers = {}

            if counter[0] > n_invalid:
                return (200, headers, payload)
            else:
                return (200, headers, b"")

        responses.add_callback(responses.GET, url,
                               callback=request_callback,
                               content_type='application/octet-stream')

    def _retry_common_setup(self):
        store_location = "http://acme.com/"
        filename = "nose-1.3.0-1.egg"

        path = os.path.join(_EGGINST_COMMON_DATA, filename)

        repository = Repository()
        package = RepositoryPackageMetadata.from_egg(path,
                                                     store_location=store_location)
        repository.add_package(package)

        downloader = _DownloadManager(mocked_session_factory(self.tempdir),
                                      repository)
        return path, downloader, repository

    @responses.activate
    def test_not_enough_retry(self):
        # Given
        path, downloader, repository = self._retry_common_setup()

        url = "http://acme.com/{0}".format(os.path.basename(path))
        with open(path, "rb") as fp:
            payload = fp.read()

        max_retries = 2
        self._add_failing_checksum_response(url, payload, max_retries)

        # When/Then
        action = FetchAction(path, downloader, repository,
                             max_retries=max_retries-1)
        with self.assertRaises(InvalidChecksum):
            action.execute()

    @responses.activate
    def test_retry(self):
        # Given
        path, downloader, repository = self._retry_common_setup()

        url = "http://acme.com/{0}".format(os.path.basename(path))
        with open(path, "rb") as fp:
            payload = fp.read()

        max_retries = 4
        self._add_failing_checksum_response(url, payload, max_retries)

        # When/Then
        action = FetchAction(path, downloader, repository,
                             max_retries=max_retries)
        # No exception
        action.execute()


class TestRemoveAction(unittest.TestCase):
    def setUp(self):
        self.top_prefix = tempfile.mkdtemp()
        self.top_installed_repository = Repository()
        self.installed_repository = Repository()

    def tearDown(self):
        shutil.rmtree(self.top_prefix)

    def _install_eggs(self, paths):
        repository = Repository()
        for path in paths:
            package = RepositoryPackageMetadata.from_egg(path)
            repository.add_package(package)

        for path in paths:
            action = InstallAction(path, self.top_prefix, repository,
                                   self.top_installed_repository,
                                   self.installed_repository,
                                   os.path.dirname(path))
            action.execute()

    def test_simple(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)

        metadata = PackageMetadata.from_egg(path)

        self._install_eggs([path])

        # When
        action = RemoveAction(path, self.top_prefix,
                              self.top_installed_repository,
                              self.installed_repository)
        action.execute()

        # Then
        repository = Repository._from_prefixes([self.top_prefix])
        self.assertFalse(repository.has_package(metadata))
        self.assertFalse(self.top_installed_repository.has_package(metadata))
        self.assertFalse(self.installed_repository.has_package(metadata))
        self.assertFalse(action.is_canceled)

    def test_iteration(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)

        metadata = PackageMetadata.from_egg(path)

        self._install_eggs([path])

        # When
        action = RemoveAction(path, self.top_prefix,
                              self.top_installed_repository,
                              self.installed_repository)
        for step in action:
            pass

        # Then
        repository = Repository._from_prefixes([self.top_prefix])
        self.assertFalse(repository.has_package(metadata))
        self.assertFalse(self.top_installed_repository.has_package(metadata))
        self.assertFalse(self.installed_repository.has_package(metadata))
        self.assertFalse(action.is_canceled)

    def test_iteration_cancel(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)

        metadata = PackageMetadata.from_egg(path)

        self._install_eggs([path])

        # When
        action = RemoveAction(path, self.top_prefix,
                              self.top_installed_repository,
                              self.installed_repository)
        for step in action:
            action.cancel()

        # Then
        repository = Repository._from_prefixes([self.top_prefix])
        self.assertFalse(repository.has_package(metadata))
        self.assertFalse(self.top_installed_repository.has_package(metadata))
        self.assertFalse(self.installed_repository.has_package(metadata))

        self.assertTrue(action.is_canceled)



from __future__ import absolute_import, print_function

import errno
import io
import os
import os.path
import sys
import textwrap

from egginst.vendor.six.moves import urllib

from egginst.progress import (console_progress_manager_factory,
                              dummy_progress_bar_factory)

from enstaller.auth import UserInfo
from enstaller.egg_meta import split_eggname
from enstaller.errors import MissingDependency, NoSuchPackage, NoPackageFound
from enstaller.legacy_stores import parse_index
from enstaller.repository import Repository, egg_name_to_name_version
from enstaller.requests_utils import _ResponseIterator
from enstaller.solver import Request, Requirement
from enstaller.utils import decode_json_from_buffer, prompt_yes_no
from enstaller.vendor.futures import ThreadPoolExecutor, as_completed


FMT = '%-20s %-20s %s'
FMT4 = '%-20s %-20s %-20s %s'

DEFAULT_TEXT_WIDTH = 79


def disp_store_info(store_location):
    if not store_location:
        return '-'
    for rm in 'http://', 'https://', 'www', '.enthought.com', '/repo/':
        store_location = store_location.replace(rm, '')
    return store_location.replace('/eggs/', ' ').strip('/')


def _is_any_package_unavailable(remote_repository, actions):
    unavailables = []
    for opcode, egg in actions:
        if opcode == "install":
            name, version = egg_name_to_name_version(egg)
            package = remote_repository.find_package(name, version)
            if not package.available:
                unavailables.append(egg)
    return len(unavailables) > 0


def _notify_unavailable_package(config, requirement, session):
    logged_in_string = config.auth.logged_message
    user_info = UserInfo.from_session(session)
    subscription = user_info.subscription_level
    msg = textwrap.dedent("""\
        Cannot install {0!r}, as this package (or some of its requirements)
        are not available at your subscription level {1!r}
        ({2}).
        """.format(str(requirement), subscription, logged_in_string))
    print()
    print(textwrap.fill(msg, DEFAULT_TEXT_WIDTH))


def _requirement_from_pypi(request, repository):
    are_pypi = []
    for job in request.jobs:
        if job.kind in ("install", "update", "upgrade"):
            try:
                candidate = \
                    repository.find_package_from_requirement(job.requirement)
            except NoSuchPackage:
                pass
            else:
                if candidate.product == "pypi":
                    are_pypi.append(job.requirement)
    return are_pypi


_BROKEN_PYPI_TEMPLATE = """
Broken pypi package '{requested}': missing dependency '{dependency}'

Pypi packages are not officially supported. If this package is important to
you, please contact Enthought support to request its inclusion in our
officially supported repository.

In the mean time, you may want to try installing '{requested}' from sources
with pip as follows:

    $ enpkg pip
    $ pip install <requested_package>
"""


def install_req(enpkg, config, req, opts):
    """
    Try to execute the install actions.
    """
    # Unix exit-status codes
    FAILURE = 1
    req = Requirement.from_anything(req)
    request = Request()
    request.install(req)

    def _done(exit_status):
        sys.exit(exit_status)

    def _get_unsupported_packages(actions):
        ret = []
        for opcode, egg in actions:
            if opcode == "install":
                name, version = egg_name_to_name_version(egg)
                package = enpkg._remote_repository.find_package(name, version)
                if package.product == "pypi":
                    ret.append(package)
        return ret

    def _ask_pypi_confirmation(package_list_string):
        msg = textwrap.dedent("""\
        The following packages/requirements are coming from the PyPi repo:

        {0}

        The PyPi repository which contains >10,000 untested ("as is")
        packages. Some packages are licensed under GPL or other licenses
        which are prohibited for some users. Dependencies may not be
        provided. If you need an updated version or if the installation
        fails due to unmet dependencies, the Knowledge Base article
        Installing external packages into Canopy Python
        (https://support.enthought.com/entries/23389761) may help you with
        installing it.
        """.format(package_list_string))
        print(msg)

        msg = "Are you sure that you wish to proceed?  (y/[n])"
        if not prompt_yes_no(msg, opts.yes):
            sys.exit(0)

    def _ask_pypi_confirmation_from_actions(actions):
        unsupported_packages = _get_unsupported_packages(actions)
        if len(unsupported_packages) > 0:
            package_list = sorted("'{0}-{1}'".format(p.name, p.full_version)
                                  for p in unsupported_packages)
            package_list_string = "\n".join(package_list)
            _ask_pypi_confirmation(package_list_string)

    try:
        mode = 'root' if opts.no_deps else 'recur'
        pypi_asked = False
        solver = enpkg._solver_factory(mode, opts.force, opts.forceall)

        pypi_requirements = _requirement_from_pypi(request,
                                                   enpkg._remote_repository)

        try:
            actions = solver.resolve(request)
        except MissingDependency as e:
            if len(pypi_requirements) > 0:
                msg = _BROKEN_PYPI_TEMPLATE.format(requested=e.requester,
                                                   dependency=e.requirement)
                print(msg)
            else:
                print("One of the requested package has broken dependencies")
                print("(Dependency solving error: {0})".format(e))
            _done(FAILURE)

        if len(pypi_requirements) > 0:
            package_list = sorted(str(p) for p in pypi_requirements)
            _ask_pypi_confirmation("\n".join(package_list))
            pypi_asked = True

        installed = (egg for opcode, egg in actions if opcode == "install")
        actions = [("fetch", egg) for egg in installed] + actions

        if _is_any_package_unavailable(enpkg._remote_repository, actions):
            _notify_unavailable_package(config, req, enpkg._session)
            _done(FAILURE)
        if not pypi_asked:
            _ask_pypi_confirmation_from_actions(actions)
        enpkg.execute(actions)
        if len(actions) == 0:
            print("No update necessary, %r is up-to-date." % req.name)
            print(install_time_string(enpkg._installed_repository,
                                      req.name))
    except NoPackageFound as e:
        print(str(e))
        _done(FAILURE)
    except OSError as e:
        if e.errno == errno.EACCES and sys.platform == 'darwin':
            print("Install failed. OSX install requires admin privileges.")
            print("You should add 'sudo ' before the 'enpkg' command.")
            _done(FAILURE)
        else:
            raise


def install_time_string(installed_repository, name):
    lines = []
    for info in installed_repository.find_packages(name):
        lines.append('%s was installed on: %s' % (info.key, info.ctime))
    return "\n".join(lines)


def print_installed(repository, pat=None):
    print(FMT % ('Name', 'Version', 'Store'))
    print(60 * '=')
    for package in repository.iter_packages():
        if pat and not pat.search(package.name):
            continue
        print(FMT % (package.name, package.full_version,
                     disp_store_info(package.store_location)))


def _should_raise(resp, raise_on_error):
    if not raise_on_error:
        if resp.status_code in (403, 404):
            return False
    return True


def _print_warning(msg, width=DEFAULT_TEXT_WIDTH):
    preambule = "Warning: "
    wrapper = textwrap.TextWrapper(initial_indent=preambule,
                                   subsequent_indent=len(preambule) * " ",
                                   width=width)
    print(wrapper.fill(msg) + "\n")


def _fetch_repository(session, url, store_location, raise_on_error):
    with session.etag():
        resp = session.get(url, stream=True)
        if resp.status_code != 200:
            if _should_raise(resp, raise_on_error):
                resp.raise_for_status()
            else:
                return None  # failed.append(store_location)
        else:
            data = io.BytesIO()
            for chunk in _ResponseIterator(resp):
                data.write(chunk)
            json_data = decode_json_from_buffer(data.getvalue())
            return Repository(parse_index(json_data, store_location))


def _print_unavailables_warning(unavailables):
    store_names = [_display_store_name(store_location) for store_location in
                   unavailables]
    preambule = "Warning: "
    template = textwrap.dedent("""\
        {0}Could not fetch the following indices:

        {1}
    """)
    displayed = "\n".join("{0}- {1!r}".format(" " * len(preambule), name)
                          for name in store_names)

    after = ("Those repositories do not exist (or you do not have the "
             "rights to access them). You should edit your configuration "
             "to remove those repositories.")
    print(template.format(preambule, displayed))
    wrapper = textwrap.TextWrapper(initial_indent=len(preambule) * " ",
                                   subsequent_indent=len(preambule) * " ",
                                   width=DEFAULT_TEXT_WIDTH)
    print(wrapper.fill(after))
    print()


def _write_and_flush(s, quiet):
    if not quiet:
        sys.stdout.write(s)
        sys.stdout.flush()


def repository_factory(session, indices, quiet=False, raise_on_error=False):
    unavailables = []
    full_repository = Repository()

    _write_and_flush("Fetching indices: ", quiet)

    with ThreadPoolExecutor(max_workers=4) as executor:
        tasks = []
        for url, store_location in indices:
            task = executor.submit(_fetch_repository, session, url,
                                   store_location, raise_on_error)
            tasks.append(task)

        for task in as_completed(tasks):
            repository_or_none = task.result()
            if repository_or_none is None:
                unavailables.append(store_location)
            else:
                for package in repository_or_none.iter_packages():
                    full_repository.add_package(package)
            _write_and_flush(".", quiet)

    _write_and_flush("\n\n", quiet)

    if len(unavailables) > 0 and not quiet:
        _print_unavailables_warning(unavailables)
    return full_repository


def name_egg(egg):
    return split_eggname(egg)[0]


def updates_check(remote_repository, installed_repository):
    updates = []
    EPD_update = []
    for package in installed_repository.iter_packages():
        av_metadatas = remote_repository.find_sorted_packages(package.name)
        if len(av_metadatas) == 0:
            continue
        av_metadata = av_metadatas[-1]
        if av_metadata.version > package.version:
            if package.name == "epd":
                EPD_update.append({'current': package, 'update': av_metadata})
            else:
                updates.append({'current': package, 'update': av_metadata})
    return updates, EPD_update


def humanize_ssl_error_and_die(ssl_exception, store_url):
    if ssl_exception.request is not None:
        url = ssl_exception.request.url
    else:
        url = store_url
    p = urllib.parse.urlparse(url)
    print("SSL error: {0}".format(str(ssl_exception)))
    print("To connect to {0!r} insecurely, add the `-k` flag to enpkg "
          "command".format(p.hostname))
    sys.exit(-1)


def is_running_on_non_owned_python():
    """Returns True if the running python process is owned by root
    and sys.executable is owned by a different uid.
    """
    if sys.platform == 'win32':
        return False
    else:
        return (os.getuid() == 0
                and os.stat(sys.executable).st_uid != os.getuid())


def exit_if_root_on_non_owned(force_yes=False):
    if is_running_on_non_owned_python():
        msg = ("You are running enpkg in a python installation not "
               "owned by root, are you sure to continue ? (y/[n])")
        if not prompt_yes_no(msg, force_yes=force_yes):
            sys.exit(-1)


# Private functions
def _fetch_json_with_progress(resp, store_location, quiet=False):
    data = io.BytesIO()

    length = int(resp.headers.get("content-length", 0))
    display = _display_store_name(store_location)
    if quiet:
        progress = dummy_progress_bar_factory()
    else:
        progress = console_progress_manager_factory("Fetching index", display,
                                                    size=length)
    with progress:
        for chunk in _ResponseIterator(resp):
            data.write(chunk)
            progress.update(len(chunk))

    data = data.getvalue()
    return decode_json_from_buffer(data)


def _display_store_name(store_location):
    parts = urllib.parse.urlsplit(store_location)
    return urllib.parse.urlunsplit(("", "", parts[2], parts[3], parts[4]))



import os.path

from egginst._compat import PY2
from egginst.vendor.six import string_types

from enstaller.auth import APITokenAuth, UserPasswordAuth
from enstaller.errors import InvalidConfiguration
from enstaller.plat import custom_plat
from enstaller.vendor import jsonschema

if PY2:
    from enstaller.vendor import yaml
else:
    from enstaller.vendor import yaml_py3 as yaml


_API_TOKEN = "api_token"
_AUTHENTICATION = "authentication"
_AUTHENTICATION_TYPE = "kind"
_AUTHENTICATION_TYPE_BASIC = "basic"
_AUTHENTICATION_TYPE_SIMPLE = "simple"
_AUTHENTICATION_TYPE_TOKEN = "token"
_MAX_RETRIES = "max_retries"
_SSL_VERIFY = "verify_ssl"
_USERNAME = "username"
_PASSWORD = "password"
_AUTH_STRING = "auth"
_REPOSITORIES = "repositories"
_FILES_CACHE = "files_cache"
_STORE_URL = "store_url"

_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "EnstallerConfiguration",
    "description": "Enstaller >= 4.8.0 configuration",
    "type": "object",
    "properties": {
        "max_retries": {
            "description": "Max number of time to retry connecting to a "
                           "remote server or re-fetching data with invalid "
                           "checksum",
            "type": "integer"
        },
        "verify_ssl": {
            "description": "Whether to actually check SSL CA certificate or "
                           "not",
            "type": "boolean"
        },
        "store_url": {
            "description": "The url (schema + hostname only of the store to "
                           "connect to).",
            "type": "string"
        },
        "files_cache": {
            "description": "Where to cache downloaded files.",
            "type": "string"
        },
        "authentication": {
            "type": "object",
            "oneOf": [
                {"$ref": "#/definitions/api_token_authentication"},
                {"$ref": "#/definitions/simple_authentication"},
                {"$ref": "#/definitions/basic_authentication"}
            ],
            "description": "Authentication."
        },
        "repositories": {
            "type": "array",
            "items": {"type": "string"},
            "description": "List of repositories."
        }
    },
    "definitions": {
        "api_token_authentication": {
            "properties": {
                "kind": {
                    "enum": [_AUTHENTICATION_TYPE_TOKEN],
                    "default": _AUTHENTICATION_TYPE_TOKEN
                },
                _API_TOKEN: {"type": "string"}
            },
            "required": [_API_TOKEN],
            "additionalProperties": False
        },
        "simple_authentication": {
            "properties": {
                "kind": {
                    "enum": [_AUTHENTICATION_TYPE_SIMPLE]
                },
                _USERNAME: {"type": "string"},
                _PASSWORD: {"type": "string"}
            },
            "required": ["kind", _USERNAME, _PASSWORD],
            "additionalProperties": False
        },
        "basic_authentication": {
            "properties": {
                "kind": {
                    "enum": [_AUTHENTICATION_TYPE_BASIC]
                },
                _AUTH_STRING: {"type": "string"}
            },
            "required": ["kind", _AUTH_STRING],
            "additionalProperties": False
        }
    },
    "additionalProperties": False,
}


def load_configuration_from_yaml(cls, filename_or_fp):
    # FIXME: local import to workaround circular import
    from enstaller.config import STORE_KIND_BROOD
    if isinstance(filename_or_fp, string_types):
        with open(filename_or_fp, "rt") as fp:
            data = yaml.load(fp)
    else:
        data = yaml.load(filename_or_fp)

    if data is None:
        data = {}
    else:
        try:
            jsonschema.validate(data, _SCHEMA)
        except jsonschema.ValidationError as e:
            msg = "Invalid configuration: {0!r}".format(e.message)
            raise InvalidConfiguration(msg)

    config = cls()

    if _AUTHENTICATION in data:
        authentication = data[_AUTHENTICATION]
        authentication_type = authentication.get(_AUTHENTICATION_TYPE,
                                                 _AUTHENTICATION_TYPE_TOKEN)
        if authentication_type == _AUTHENTICATION_TYPE_SIMPLE:
            username = authentication[_USERNAME]
            password = authentication[_PASSWORD]
            auth = UserPasswordAuth(username, password)
        elif authentication_type == _AUTHENTICATION_TYPE_BASIC:
            auth_string = authentication[_AUTH_STRING]
            auth = UserPasswordAuth.from_encoded_auth(auth_string)
        elif authentication_type == _AUTHENTICATION_TYPE_TOKEN:
            token = authentication[_API_TOKEN]
            auth = APITokenAuth(token)
        else:
            msg = "Unknown authentication type {0!r}". \
                  format(authentication_type)
            raise InvalidConfiguration(msg)
        config.update(auth=auth)

    if _STORE_URL in data:
        config.update(store_url=data[_STORE_URL])
    if _REPOSITORIES in data:
        config.set_repositories_from_names(data[_REPOSITORIES])

    if _FILES_CACHE in data:
        files_cache = os.path.expanduser(data[_FILES_CACHE]). \
            replace("{PLATFORM}", custom_plat)
        config._repository_cache = files_cache
    if _MAX_RETRIES in data:
        config.update(max_retries=data[_MAX_RETRIES])
    if _SSL_VERIFY in data and not data[_SSL_VERIFY]:
        config.update(verify_ssl=data[_SSL_VERIFY])

    config.update(use_webservice=False)

    if isinstance(filename_or_fp, string_types):
        config._filename = filename_or_fp

    config._store_kind = STORE_KIND_BROOD
    return config



import hashlib
import contextlib

from egginst.utils import atomic_file
from enstaller.errors import InvalidChecksum


class MD5File(object):
    def __init__(self, fp):
        """
        A simple file object wrapper that computes a md5 checksum only when data
        are being written

        Parameters
        ----------
        fp: file object-like
            The file object to wrap.
        """
        self._fp = fp
        self._h = hashlib.md5()
        self._aborted = False

    @property
    def is_aborted(self):
        return self._aborted

    @property
    def checksum(self):
        return self._h.hexdigest()

    def abort(self):
        self._aborted = True

    def write(self, data):
        """
        Write the given data buffer to the underlying file.
        """
        self._fp.write(data)
        self._h.update(data)


@contextlib.contextmanager
def checked_content(filename, expected_md5):
    """
    A simple context manager ensure data written to filename match the given
    md5.

    Parameters
    ----------
    filename : str
        The path to write to
    expected_checksum : str
        The expected checksum

    Returns
    -------
    fp : MD5File
        A file-like MD5File instance.

    Example
    -------
    A simple example::

        with checked_content("foo.bin", expected_md5) as fp:
            fp.write(data)
        # An InvalidChecksum will be raised if the checksum does not match
        # expected_md5

    The checksum may be disabled by setting up abort to fp::

        with checked_content("foo.bin", expected_md5) as fp:
            fp.write(data)
            fp.abort = True
            # no checksum is getting validated
    """
    with atomic_file(filename) as target:
        checked_target = MD5File(target)
        yield checked_target

        if checked_target.is_aborted:
            target.abort()
            return
        else:
            if expected_md5 != checked_target.checksum:
                raise InvalidChecksum(filename, expected_md5,
                                      checked_target.checksum)



from __future__ import absolute_import

import os
import sys
import time


from ._termui import get_terminal_size


if os.name == 'nt':
    BEFORE_BAR = '\r'
    AFTER_BAR = '\n'
else:
    BEFORE_BAR = '\r\033[?25l'
    AFTER_BAR = '\033[?25h\n'


_KIB = 1024


class _DummyStream(object):
    def flush(self):
        pass

    def write(self, data):
        pass


class ExpAverager(object):
    def __init__(self, fill_size=3):
        self.fill_size = fill_size
        self._cur = None
        self.alpha = 0.9
        self._hit = 0

    @property
    def is_filled(self):
        return self._hit >= self.fill_size

    def append(self, value):
        if self._cur is None:
            self._cur = value
        else:
            self._cur = self.alpha * self._cur + (1 - self.alpha) * value

        self._hit += 1

    @property
    def value(self):
        return self._cur


def _human_speed(speed):
    if speed < 1.:
        return "-- b/sec"
    else:
        if speed > _KIB ** 2:
            speed = speed / _KIB ** 2
            unit = "MiB"
        elif speed > _KIB:
            speed = speed / _KIB
            unit = "KiB"
        else:
            unit = "b"
        return "%.1f %s/sec" % (speed, unit)


_MAX_SPEED_LABEL_DISPLAY = len(_human_speed((_KIB - 1) * _KIB))


class ProgressBar(object):
    def __init__(self, length=None, bar_template='%(bar)s', width=30,
                 fill_char='#', show_speed=False):
        if sys.stdout.isatty():
            self._stream = sys.stdout
        else:
            self._stream = _DummyStream()

        self._entered = False
        self._pos = 0
        self.fill_char = fill_char

        if length is not None:
            self._length = length
        else:
            raise ValueError("no length not supported yet.")

        self.width = width
        self.last_written_bar_pos = 0
        self.bar_template = bar_template

        self.show_speed = show_speed

        self._start = self._last_render = time.time()
        self._avg_speed = ExpAverager(5)
        self._elapsed = 0.

    def render_finish(self):
        self._stream.write(AFTER_BAR)
        self._stream.flush()

    @property
    def _elapsed_since_last_render(self):
        return time.time() - self._last_render

    @property
    def has_length(self):
        return self._length is not None

    @property
    def needs_render(self):
        # We force a render at least 4x / sec
        return (self.relative_bar_pos > self.last_written_bar_pos
                or self._elapsed_since_last_render >= 0.25) \
            and self.relative_bar_pos <= self.width

    @property
    def relative_bar_pos(self):
        assert self.has_length
        if self._length == 0:
            return 0
        else:
            return int(self._pos * 1. / self._length * self.width)

    @property
    def speed(self):
        if self._elapsed >= 0.1:
            return self._pos / self._elapsed
        else:
            return 0.

    def _update_time_state(self):
        now = time.time()
        self._last_render = now

        self._elapsed = now - self._start

        self._last_pos = self._pos
        self._last = now

    def update(self, n_steps):
        self._pos += n_steps

        if self.needs_render:
            self._update_time_state()

            if self.show_speed:
                speed = self.speed

                self._avg_speed.append(speed)
                if self._avg_speed.is_filled:
                    speed = self._avg_speed.value

                self.info = _human_speed(speed)
            else:
                self.info = ""

            self._render()

    def _render(self):
        self.last_written_bar_pos = self.relative_bar_pos
        bar_value = self.fill_char * self.relative_bar_pos + \
            " " * (self.width - self.relative_bar_pos)
        data = {
            "bar": bar_value,
            "info": self.info,
            "label": "",
        }
        rendered_bar = self.bar_template % data
        bar = BEFORE_BAR + rendered_bar
        self._stream.write(bar)
        self._stream.write(" " * (get_terminal_size()[0] - len(bar)))
        self._stream.flush()

    def __enter__(self):
        self._entered = True
        return self

    def __exit__(self, *a, **kw):
        self.render_finish()

    def __iter__(self):
        if not self._entered:
            msg = "You need to use progress bars in a `with` block."
            raise RuntimeError(msg)
        return self
